import operator
import time
from pathlib import Path

from lxml import etree

from tests.utils import delayed_assertion
from wkz import models
from wkz.demo import copy_demo_fit_files_to_track_dir
from wkz.watchdogs import trigger_device_watchdog, trigger_file_watchdog


def test__start_file_importer_watchdog_basic(db, tmp_path, test_data_dir, demo_data_dir, fit_file_a):
    assert models.Activity.objects.count() == 0
    assert models.BestSection.objects.count() == 0

    # update path_to_trace_dir in db accordingly, since file importer will read it from the db
    settings = models.get_settings()
    trace_dir = tmp_path / "trace_dir"
    trace_dir.mkdir()
    assert trace_dir.is_dir()
    settings.path_to_trace_dir = trace_dir
    settings.save()

    # put an activity fit file into the watched dir
    copy_demo_fit_files_to_track_dir(
        source_dir=demo_data_dir,
        targe_dir=trace_dir,
        list_of_files_to_copy=[fit_file_a],
    )
    assert (Path(trace_dir) / fit_file_a).is_file()

    trigger_file_watchdog()
    # watchdog should now have triggered the file imported and activity should be in db
    delayed_assertion(models.Activity.objects.count, operator.eq, 1)
    delayed_assertion(models.BestSection.objects.count, operator.gt, 0)
    bs1 = models.BestSection.objects.count()

    # now put a activity GPX file into the watched dir
    copy_demo_fit_files_to_track_dir(
        source_dir=test_data_dir,
        targe_dir=trace_dir,
        list_of_files_to_copy=["example.gpx"],
    )
    path_to_gpx = Path(trace_dir) / "example.gpx"
    assert path_to_gpx.is_file()
    # ensure file is a well-formed xml
    etree.parse(str(path_to_gpx))

    trigger_file_watchdog()

    delayed_assertion(models.Activity.objects.count, operator.eq, 2)
    delayed_assertion(models.BestSection.objects.count, operator.gt, bs1)
    bs2 = models.BestSection.objects.count()

    # create a non fit/gpx file to verify it won't be imported
    dummy_file = trace_dir / "fake_activity.txt"
    dummy_file.write_text("not a real activity", encoding="utf-8")

    # check that the dummy file was actually created
    assert Path(dummy_file).is_file()

    # but assert that the number of activities and best sections did not increase
    delayed_assertion(models.Activity.objects.count, operator.eq, 2)
    delayed_assertion(models.BestSection.objects.count, operator.eq, bs2)


def test_fake_device(db, fake_device, device_dir, activity_dir, fit_file):
    # initialize fake device
    device = fake_device(activity_files=[fit_file])

    settings = models.Settings.objects.get()
    mount_path = Path(settings.path_to_garmin_device)
    # once fake device is initialized, mount path should be available
    assert mount_path.is_dir()

    # now mount the fake device, which should also contain a fit file
    device.mount()

    assert (mount_path / device_dir).is_dir()
    assert (mount_path / device_dir / activity_dir).is_dir()
    assert (mount_path / device_dir / activity_dir / fit_file).is_file()

    # check that mounting again does not have any effect
    device.mount()
    assert (mount_path / device_dir / activity_dir / fit_file).is_file()

    # unmount device again and verify dirs are gone
    device.unmount()
    assert not (mount_path / device_dir).is_dir()
    assert not (mount_path / device_dir / activity_dir / fit_file).is_file()

    # check that unmounting again does not have any effect
    device.unmount()
    assert not (mount_path / device_dir).is_dir()
    assert not (mount_path / device_dir / activity_dir / fit_file).is_file()

    # and again mounting should also work
    device.mount()
    assert (mount_path / device_dir / activity_dir / fit_file).is_file()


def test__start_device_watchdog__missing_dir(db, caplog):
    invalid_dir = "/some/random/non_existent/path/"

    settings = models.get_settings()
    settings.path_to_garmin_device = invalid_dir
    settings.save()

    trigger_device_watchdog()
    assert f"Device Watchdog: {invalid_dir} is not a valid directory." in caplog.text


def test__start_file_importer_watchdog__missing_dir(db, caplog):
    invalid_dir = "/some/random/non_existent/path/"

    settings = models.get_settings()
    settings.path_to_trace_dir = invalid_dir
    settings.save()

    trigger_file_watchdog()
    assert f"File Watchdog: {invalid_dir} is not a valid directory." in caplog.text


def test__start_device_watchdog__collect_files(
    db, fake_device, device_dir, activity_dir, fit_file_a, fit_file_b, tmp_path
):
    # initialize fake device with two fit files
    device = fake_device(activity_files=[fit_file_a, fit_file_b])

    settings = models.get_settings()
    mount_path = Path(settings.path_to_garmin_device)
    trace_dir = tmp_path / "random_trace_dir"
    trace_dir.mkdir()
    settings.path_to_trace_dir = trace_dir
    settings.save()

    # verify mount path exists
    assert mount_path.is_dir()

    # ensure fit files are not already present in trace dir
    assert not (trace_dir / "garmin" / fit_file_a).is_file()
    assert not (trace_dir / "garmin" / fit_file_b).is_file()

    # now mount device which contains fit files
    device.mount()

    # start device watch dog
    trigger_device_watchdog()

    assert (device.activity_path_on_device / fit_file_a).is_file()
    assert (device.activity_path_on_device / fit_file_b).is_file()

    # verify the fit files are present on the mounted device
    assert (mount_path / device_dir / activity_dir / fit_file_a).is_file()
    assert (mount_path / device_dir / activity_dir / fit_file_b).is_file()

    # verify that the fit files are now present in the trace dir
    delayed_assertion((trace_dir / "garmin" / fit_file_a).is_file, operator.is_, True)
    delayed_assertion((trace_dir / "garmin" / fit_file_b).is_file, operator.is_, True)


def test_device_and_file_importer_watchdog(
    db, tmpdir, test_data_dir, demo_data_dir, fake_device, device_dir, activity_dir, fit_file_a, fit_file_b
):
    assert models.Activity.objects.count() == 0
    assert models.BestSection.objects.count() == 0

    device = fake_device(activity_files=[fit_file_a, fit_file_b])

    settings = models.get_settings()
    mount_path = Path(settings.path_to_garmin_device)
    trace_dir = Path(settings.path_to_trace_dir)
    settings.path_to_trace_dir = trace_dir
    settings.save()

    # verify mount path exists
    assert mount_path.is_dir()

    # ensure fit files are not already present in trace dir
    assert not (trace_dir / fit_file_a).is_file()
    assert not (trace_dir / fit_file_b).is_file()

    # mounting the device will:
    #  1. trigger device watchdog to copy fit files to trace dir, what in turn will
    #  2. trigger file importer watchdog to import the fit files into workoutizer activities
    device.mount()

    # start device watchdog
    trigger_device_watchdog()

    # verify the fit files are present on the mounted device
    assert (mount_path / device_dir / activity_dir / fit_file_a).is_file()
    assert (mount_path / device_dir / activity_dir / fit_file_b).is_file()

    # verify that the fit files are now present in the trace dir
    delayed_assertion((trace_dir / "garmin" / fit_file_a).is_file, operator.is_, True)
    delayed_assertion((trace_dir / "garmin" / fit_file_b).is_file, operator.is_, True)

    # start file watchdog
    trigger_file_watchdog()

    # check that the activities got imported
    delayed_assertion(models.Activity.objects.count, operator.eq, 2)
    delayed_assertion(models.BestSection.objects.count, operator.gt, 2)


def test_file_importer__with_path_being_changed(db, tmp_path, demo_data_dir, fit_file_a, fit_file_b):
    dummy_path_settings = models.Settings(path_to_trace_dir=tmp_path / "dummy_path")
    dummy_path_settings.save()
    assert models.Activity.objects.count() == 0
    assert models.BestSection.objects.count() == 0

    # assume we have two directories: an empty directory and one with activities
    empty_dir = tmp_path / "empty_dir"
    empty_dir.mkdir()
    assert empty_dir.is_dir()
    assert not any(empty_dir.iterdir())  # = no file in empty dir

    dir_with_files = tmp_path / "dir_with_files"
    dir_with_files.mkdir()
    assert dir_with_files.is_dir()
    copy_demo_fit_files_to_track_dir(
        source_dir=demo_data_dir,
        targe_dir=dir_with_files,
        list_of_files_to_copy=[fit_file_a],
    )
    assert (Path(dir_with_files) / fit_file_a).is_file()
    assert any(dir_with_files.iterdir())  # dir_with_files contains a file

    # first, file watchdog is watching the empty dir (note, chaning settings will re-trigger watchdog)
    dummy_path_settings.path_to_trace_dir = empty_dir
    dummy_path_settings.save()

    # since watchdog is watching empty dir, activity count should still be 0
    delayed_assertion(models.Activity.objects.count, operator.eq, 0)
    # now change the settings path to point to the dir_with_files
    dummy_path_settings.path_to_trace_dir = dir_with_files
    dummy_path_settings.save()

    # watchdog should automatically pick up the new path, start watching the new dir, find the
    # given activity file (even if it was added before starting to watch if) and import it
    delayed_assertion(models.Activity.objects.count, operator.eq, 1)

    # now also add an activity file to the empty dir and check that watchdog does not get triggered
    copy_demo_fit_files_to_track_dir(
        source_dir=demo_data_dir,
        targe_dir=empty_dir,
        list_of_files_to_copy=[fit_file_b],
    )
    time.sleep(3)
    delayed_assertion(models.Activity.objects.count, operator.ne, 2)



from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path

# custom error handler
handler404 = "wkz.views.custom_400_view"
handler500 = "wkz.views.custom_500_view"
handler403 = "wkz.views.custom_400_view"
handler400 = "wkz.views.custom_400_view"

urlpatterns = [
    path("admin/", admin.site.urls),
    path("", include("wkz.urls")),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



import os

from lxml import etree

from wkz.file_helper.gpx_exporter import save_activity_to_gpx_file


def test_save_activity_to_gpx_file(activity):
    path = save_activity_to_gpx_file(activity)
    assert os.path.isfile(path)
    assert path.split("/")[-1] == "2020-07-07_evening-cycling-along-the-river.gpx"
    # verify xml file is well formed this would raise lxml.etree.XMLSyntaxError if not
    etree.parse(path)



import datetime
import logging
import os
from typing import List

from sportgems import (
    DistanceTooSmallException,
    InconsistentLengthException,
    NoSectionFoundException,
    TooFewDataPointsException,
)

from wkz import configuration

log = logging.getLogger(__name__)


class Parser:
    def __init__(self, path_to_file: str, md5sum: str):
        # basic activity info
        self.path_to_file = path_to_file
        self.file_name = None
        self.md5sum = md5sum
        self.sport = None
        self.date = None
        self.duration = datetime.timedelta(minutes=0)
        self.distance = 0
        self.calories = None
        # trace file info
        # coordinates
        self.latitude_list = []
        self.longitude_list = []
        # distances
        self.distance_list = []
        # elevation
        self.altitude_list = []
        self.min_altitude = None  # float()
        self.max_altitude = None
        # heart rate
        self.heart_rate_list = []
        self.min_heart_rate = None
        self.avg_heart_rate = None  # int()
        self.max_heart_rate = None
        # cadence
        self.cadence_list = []
        self.min_cadence = None
        self.avg_cadence = None  # int()
        self.max_cadence = None
        # speed
        self.speed_list = []
        self.min_speed = None
        self.avg_speed = None  # float()
        self.max_speed = None
        # temperature
        self.temperature_list = []
        self.min_temperature = None
        self.avg_temperature = None  # int()
        self.max_temperature = None
        # training effect
        self.aerobic_training_effect = None  # float()
        self.anaerobic_training_effect = None  # float()
        # total ascent/descent
        self.total_ascent = None  # int
        self.total_descent = None  # int
        # timestamps
        self.timestamps_list = []
        # lists
        self.list_attributes = [
            self.timestamps_list,
            self.longitude_list,
            self.latitude_list,
            self.temperature_list,
            self.cadence_list,
            self.altitude_list,
            self.speed_list,
            self.heart_rate_list,
        ]
        # lap info
        self.laps = []

        # best sections
        self.best_sections = []

        # run parser
        self._parse_metadata()
        self._parse_records()
        self._post_process_data()

    def _parse_metadata(self):
        raise NotImplementedError

    def _parse_records(self):
        raise NotImplementedError

    def _post_process_data(self):
        raise NotImplementedError

    def get_file_created_datetime(self):
        self.date = datetime.datetime.fromtimestamp(os.path.getctime(self.path_to_file)).date()

    def get_file_name_from_path(self, path: str):
        return path.split("/")[-1]

    def get_best_sections(self):
        log.debug("parsing best sections using sportgems...")
        # helper func to be called for each available section kind parser

        def _get_best_sections_for_section_kind(section_parser, section_distances: List[int]):
            for distance in section_distances:
                if self.distance * 1000 > distance and self.latitude_list:
                    try:
                        result = section_parser(distance, self)
                        if result:
                            self.best_sections.append(result)
                    except (
                        DistanceTooSmallException,
                        TooFewDataPointsException,
                        NoSectionFoundException,
                        InconsistentLengthException,
                    ) as e:
                        # catching some of the sportgems customs exceptions and logging it
                        log.debug(f"Could not find requested section. Sportgems error: {e}")
                        # however some are not caught and should actually be raised,
                        # e.g NoSectionFoundException and InvalidDesiredDistanceException

        for bs in configuration.best_sections:
            _get_best_sections_for_section_kind(bs["parser"], bs["distances"])



import os

console_format = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
minimal_format = "%(message)s"


def _get_formatter_and_handler(path_to_log_dir: str, use_minimal_format: bool = False):
    if not os.path.isdir(path_to_log_dir):
        os.mkdir(path_to_log_dir)
    logging_dict = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "colored": {
                "()": "coloredlogs.ColoredFormatter",
                "format": minimal_format if use_minimal_format else console_format,
                "datefmt": "%m-%d %H:%M:%S",
            },
            "format_for_file": {
                "format": "%(asctime)s :: %(levelname)s :: %(funcName)s in %(filename)s (l:%(lineno)d) :: %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "colored",
            },
            "file": {
                "level": "WARNING",
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "format_for_file",
                "filename": os.path.join(path_to_log_dir, "wkz.log"),
                "maxBytes": 5_000_000,
                "backupCount": 5,
            },
        },
        "loggers": {
            "": {
                "level": "DEBUG",
                "handlers": ["console", "file"],
            },
        },
    }
    return logging_dict


def get_logging_config(django_log_level: str, wkz_log_level: str, huey_log_level: str, path_to_log_dir: str):
    logging_dict = _get_formatter_and_handler(path_to_log_dir=path_to_log_dir)
    logging_dict["loggers"] = {
        "django": {
            "handlers": ["console", "file"],
            "level": django_log_level,
        },
        "wkz": {
            "handlers": ["console", "file"],
            "level": wkz_log_level,
        },
        "huey": {
            "handlers": ["console", "file"],
            "level": huey_log_level,
        },
    }

    return logging_dict



import datetime
import os

import pytz

from wkz import models
from wkz.demo import (
    change_date_of_demo_activities,
    copy_demo_fit_files_to_track_dir,
    insert_custom_demo_activities,
    insert_demo_sports_to_model,
)
from workoutizer import settings as django_settings


def test_insert_demo_sports_to_model(db, flush_db):
    # assert that there are no sports
    assert models.Sport.objects.count() == 0
    # insert sports
    insert_demo_sports_to_model(models)
    # check that all sports are created
    assert models.Sport.objects.count() == 5


def test_copy_demo_fit_files_to_track_dir__all(tmpdir, demo_data_dir):
    # copy demo data to tested data dir
    src = demo_data_dir
    copy_demo_fit_files_to_track_dir(str(src), str(tmpdir), list_of_files_to_copy=[])
    assert os.path.isfile(os.path.join(tmpdir, "hike_with_coordinates_muggenbrunn.fit"))
    number_of_files_in_dir = len([name for name in os.listdir(tmpdir) if os.path.isfile(os.path.join(tmpdir, name))])
    assert number_of_files_in_dir == 10


def test_copy_demo_fit_files_to_track_dir__not_all(tmpdir, demo_data_dir):
    # copy demo data to tested data dir
    src = demo_data_dir
    copy_demo_fit_files_to_track_dir(
        src,
        tmpdir,
        list_of_files_to_copy=[
            "hike_with_coordinates_muggenbrunn.fit",
            "2020-08-20-09-34-33.fit",
        ],
    )
    assert os.path.isfile(os.path.join(tmpdir, "hike_with_coordinates_muggenbrunn.fit"))
    assert os.path.isfile(os.path.join(tmpdir, "2020-08-20-09-34-33.fit"))
    number_of_files_in_dir = len([name for name in os.listdir(tmpdir) if os.path.isfile(os.path.join(tmpdir, name))])
    assert number_of_files_in_dir == 2


def test_change_date_of_demo_activities(db, sport):
    # add Activities to test the changed date
    activity_a = models.Activity(
        name="activity_a",
        sport=sport,
        date=datetime.datetime(2000, 1, 1, tzinfo=pytz.utc),
        duration=datetime.timedelta(hours=1),
        distance=10.0,
        is_demo_activity=True,
    )
    activity_a.save()
    activity_b = models.Activity(
        name="activity_b",
        sport=sport,
        date=datetime.datetime(2000, 1, 1, tzinfo=pytz.utc),
        duration=datetime.timedelta(hours=1),
        distance=10.0,
        is_demo_activity=True,
    )
    activity_b.save()
    activity_c = models.Activity(
        name="activity_c",
        sport=sport,
        date=datetime.datetime(2000, 1, 1, tzinfo=pytz.utc),
        duration=datetime.timedelta(hours=1),
        distance=10.0,
        is_demo_activity=False,  # no demo activity
    )
    activity_c.save()
    demo_activities = models.Activity.objects.all()
    change_date_of_demo_activities(every_nth_day=3, activities=demo_activities)

    a = models.Activity.objects.get(name="activity_a")
    b = models.Activity.objects.get(name="activity_b")
    c = models.Activity.objects.get(name="activity_c")

    # the date of activity a and b will be shifted to today - 1 or today - 3
    assert (
        datetime.datetime.now(pytz.timezone(django_settings.TIME_ZONE)).date() - datetime.timedelta(days=1)
        == a.date.date()
    )
    assert (
        datetime.datetime.now(pytz.timezone(django_settings.TIME_ZONE)).date() - datetime.timedelta(days=4)
        == b.date.date()
    )

    # the date of c should not be shifted
    assert datetime.datetime(2000, 1, 1, tzinfo=pytz.utc) == c.date


def test_insert_custom_demo_activities(db):
    activities = models.Activity.objects.all()
    assert len(activities) == 0
    insert_custom_demo_activities(
        count=8,
        every_nth_day=4,
        activity_model=models.Activity,
        sport_model=models.Sport,
    )
    activities = models.Activity.objects.all()
    assert len(activities) == 8



import shutil
from pathlib import Path

from wkz.file_helper.fit_collector import FitCollector, _find_activity_sub_dir_in_path


def test__find_activity_sub_dir_in_path(tmp_path):

    # desired sub dir is in first level
    d = tmp_path / "first"
    d.mkdir()
    assert _find_activity_sub_dir_in_path(name_of_dir="first", path=tmp_path, depth=1) == str(d)

    # desired sub dir is in first level and algorithm should not be case sensitive
    d = tmp_path / "UPPER"
    d.mkdir()
    assert _find_activity_sub_dir_in_path(name_of_dir="upper", path=tmp_path, depth=1).lower() == str(d).lower()

    # desired sub dir is in second level
    d = tmp_path / "first" / "second"
    d.mkdir()
    assert _find_activity_sub_dir_in_path(name_of_dir="second", path=tmp_path, depth=2) == str(d)

    # desired sub dir is in third level
    d = tmp_path / "first" / "second" / "third"
    d.mkdir()
    assert _find_activity_sub_dir_in_path(name_of_dir="third", path=tmp_path, depth=3) == str(d)

    # if dir in fourth level is searched for but depth is specified as 3 only, function should return False
    d = tmp_path / "first" / "second" / "third" / "fourth"
    d.mkdir()
    assert _find_activity_sub_dir_in_path(name_of_dir="fourth", path=tmp_path, depth=3) is False

    # a more realistic example
    d = tmp_path / "mtp:something" / "GARMIN" / "Primary" / "Activity"
    d.mkdir(parents=True)
    assert _find_activity_sub_dir_in_path(name_of_dir="Activity", path=tmp_path, depth=4) == str(d)


def test_deleting_fit_files_after_coying(tmp_path, demo_data_dir):
    # path to garmin device
    garmin = tmp_path / "garmin"
    garmin.mkdir()

    # activity dir on device
    activity = garmin / "Activity"
    activity.mkdir()

    # path to target activity dir
    target = tmp_path / "target"
    target.mkdir()

    # copy file from demo data dir to activity dir "on device"
    fit_file_1 = Path(activity) / "test_fit.fit"
    source_fit_1 = Path(demo_data_dir) / "cycling_bad_schandau.fit"
    shutil.copy(source_fit_1, fit_file_1)

    assert fit_file_1.is_file()

    # first collect fit file without deleting source file
    fit_collector = FitCollector(garmin, target, delete_files_after_import=False)
    fit_collector.copy_fit_files()

    # verify fit file got copied
    assert (target / "garmin" / "test_fit.fit").is_file()

    # file "on device" did not get deleted
    assert (activity / "test_fit.fit").is_file()

    fit_file_2 = Path(activity) / "test_fit_2.fit"
    source_fit_2 = Path(demo_data_dir) / "2020-08-28-11-57-10.fit"
    shutil.copy(source_fit_2, fit_file_2)
    assert fit_file_2.is_file()

    # now collect it with deleting it
    fit_collector = FitCollector(garmin, target, delete_files_after_import=True)
    fit_collector.copy_fit_files()

    # verify fit file got copied
    assert (target / "garmin" / "test_fit_2.fit").is_file()

    # file "on device" got deleted
    assert not (activity / "test_fit_2.fit").is_file()


def test_collecting_fit_files_with_upper_case_ending(tmp_path, demo_data_dir):
    # path to garmin device
    garmin = tmp_path / "garmin"
    garmin.mkdir()

    # activity dir on device
    activity = garmin / "Activity"
    activity.mkdir()

    # path to target activity dir
    target = tmp_path / "target"
    target.mkdir()

    # copy demo fit file and use upper case file ending
    fit_file = Path(activity) / "test_fit.FIT"
    source_fit = Path(demo_data_dir) / "cycling_bad_schandau.fit"
    shutil.copy(source_fit, fit_file)

    assert fit_file.is_file()

    fit_collector = FitCollector(garmin, target)
    fit_collector.copy_fit_files()

    # verify fit file got copied
    assert (target / "garmin" / "test_fit.FIT").is_file()



import logging
from math import pi

import pandas as pd
from bokeh.embed import components
from bokeh.plotting import figure
from bokeh.transform import cumsum

log = logging.getLogger(__name__)


def plot_pie_chart(activities):
    sport_distribution = {}
    color_list = []
    for activity in activities:
        sport_distribution[activity.sport.name] = 0
        if activity.sport.color not in color_list:
            color_list.append(activity.sport.color)
    for activity in activities:
        if activity.sport.name in sport_distribution:
            sport_distribution[activity.sport.name] += 1

    data = pd.Series(sport_distribution).reset_index(name="value").rename(columns={"index": "sport"})
    data["angle"] = data["value"] / data["value"].sum() * 2 * pi
    data["color"] = color_list

    p = figure(
        plot_height=250,
        toolbar_location=None,
        sizing_mode="scale_width",
        tools="hover",
        tooltips="@sport: @value",
        x_range=(-0.5, 1.0),
    )

    p.wedge(
        x=0.3,
        y=0.5,
        radius=0.3,
        start_angle=cumsum("angle", include_zero=True),
        end_angle=cumsum("angle"),
        line_color="white",
        fill_color="color",
        source=data,
    )

    p.axis.axis_label = None
    p.axis.visible = False
    p.grid.grid_line_color = None
    p.outline_line_color = None

    script_pc, div_pc = components(p)

    return script_pc, div_pc



import datetime
import os
import shutil
from typing import List

import pytest
import pytz
from django.core.management import call_command
from py._path.local import LocalPath

from wkz import models
from wkz.demo import copy_demo_fit_files_to_track_dir, prepare_import_of_demo_activities
from wkz.file_importer import run_importer__dask
from workoutizer import settings as django_settings


@pytest.fixture
def tracks_in_tmpdir(db, tmp_path):
    path = tmp_path / "test_traces"
    path.mkdir()
    settings = models.get_settings()
    settings.path_to_trace_dir = str(path)
    settings.save()
    return settings


@pytest.fixture
def flush_db(db):
    call_command("flush", verbosity=0, interactive=False)


@pytest.fixture
def settings(db):
    settings = models.Settings(
        path_to_trace_dir="/home/pi/traces/",
        path_to_garmin_device="/home/pi/traces/",
        number_of_days=30,
        delete_files_after_import=False,
    )
    settings.save()
    return settings


@pytest.fixture
def sport(db):
    sport = models.Sport(name="Cycling", color="red", icon="Bike")
    sport.save()
    return sport


@pytest.fixture
def trace_file(db):
    trace = models.Traces(
        path_to_file="some/path/to/file.gpx",
        file_name="file.gpx",
        md5sum="4c1185c55476269b442f424a9d80d964",
        latitude_list="[49.47972273454071, 49.47972273454071]",
        longitude_list="[8.47357001155615, 8.47357001155615]",
        calories=123,
    )
    trace.save()
    return trace


@pytest.fixture
def activity(db, sport, trace_file):
    activity = models.Activity(
        name="Evening Cycling along the River",
        sport=sport,
        date=datetime.datetime(2020, 7, 7, tzinfo=pytz.timezone(django_settings.TIME_ZONE)),
        duration=datetime.timedelta(minutes=30),
        distance=5.2,
        description="some super sport",
        trace_file=trace_file,
    )
    activity.save()
    return activity


@pytest.fixture
def import_demo_data(tracks_in_tmpdir):
    prepare_import_of_demo_activities(models)
    assert models.Sport.objects.count() == 5
    assert models.Settings.objects.count() == 1

    run_importer__dask(models, importing_demo_data=True)
    assert models.Activity.objects.count() > 1


@pytest.fixture
def import_one_activity(tracks_in_tmpdir, test_data_dir, demo_data_dir):
    models.get_settings()
    assert models.Settings.objects.count() == 1

    def _copy_activity(file_name: str):
        file_in_test_data_dir = os.path.join(test_data_dir, file_name)
        file_in_demo_data_dir = os.path.join(demo_data_dir, file_name)
        if os.path.isfile(file_in_test_data_dir):
            source_dir = test_data_dir
            path = file_in_test_data_dir
        elif os.path.isfile(file_in_demo_data_dir):
            source_dir = demo_data_dir
            path = file_in_demo_data_dir
        else:
            raise FileNotFoundError(f"file {file_name} neither found in {test_data_dir} nor in {demo_data_dir}")
        copy_demo_fit_files_to_track_dir(
            source_dir=source_dir,
            targe_dir=models.get_settings().path_to_trace_dir,
            list_of_files_to_copy=[path],
        )
        run_importer__dask(models)
        assert models.Activity.objects.count() == 1

    return _copy_activity


@pytest.fixture
def insert_sport(db):
    def _create_sport(name="Cycling", icon="bicycle"):
        sport = models.Sport(name=name, color="red", icon=icon)
        sport.save()
        return sport

    return _create_sport


@pytest.fixture
def insert_activity(db, sport):
    def _create_activity(
        name: str = "Evening Cycling along the River",
        evaluates_for_awards: bool = True,
        sport=sport,
        date: datetime.datetime = datetime.datetime(2020, 7, 7, tzinfo=pytz.timezone(django_settings.TIME_ZONE)),
        duration: datetime.timedelta = datetime.timedelta(minutes=30),
    ):
        activity = models.Activity(
            name=name,
            sport=sport,
            date=date,
            duration=duration,
            distance=5.2,
            description="some super activity",
            evaluates_for_awards=evaluates_for_awards,
        )
        activity.save()
        return activity

    return _create_activity


@pytest.fixture
def insert_best_section(db, activity, insert_activity):
    def _create_section(max_value: float):
        best_section = models.BestSection(
            activity=insert_activity(),  # always use a new activity
            kind="fastest",
            distance=1,
            start=5,
            end=517,
            max_value=max_value,
        )
        best_section.save()
        return best_section

    return _create_section


class FakeDevice:
    def __init__(self, mount_path: LocalPath, device_dir, activity_dir: str, activity_files):
        self.mount_path = mount_path
        self.device_path = self.mount_path / device_dir
        self.activity_path_on_device = self.device_path / activity_dir
        self.mounted = False
        self.activity_files = activity_files

    def mount(self):
        # create directories for activities
        if not self.mounted:
            print("mounting fake device...")
            self.activity_path_on_device.mkdir(parents=True)
            if self.activity_files:
                # copy activity files into activity dir
                copy_demo_fit_files_to_track_dir(
                    source_dir=django_settings.INITIAL_TRACE_DATA_DIR,
                    targe_dir=self.activity_path_on_device,
                    list_of_files_to_copy=self.activity_files,
                )
            self.mounted = True

    def unmount(self):
        if self.mounted:
            print("unmounting fake device...")
            shutil.rmtree(self.device_path)
            self.mounted = False


@pytest.fixture
def fake_device(tmp_path, device_dir, activity_dir):
    settings = models.get_settings()

    mount_path = tmp_path / "mount_path"
    mount_path.mkdir()
    assert mount_path.is_dir()
    settings.path_to_garmin_device = mount_path

    trace_dir = tmp_path / "trace_dir"
    trace_dir.mkdir()
    assert trace_dir.is_dir()
    settings.path_to_trace_dir = trace_dir

    settings.save()

    def _get_device(activity_files: List[str], device_dir: str = device_dir, activity_dir: str = activity_dir):
        return FakeDevice(
            mount_path=settings.path_to_garmin_device,
            device_dir=device_dir,
            activity_dir=activity_dir,
            activity_files=activity_files,
        )

    return _get_device


@pytest.fixture
def device_dir():
    return "mtp:host_DEVICE"


@pytest.fixture
def activity_dir():
    return "Primary/GARMIN/Activity"



import datetime

import pandas as pd
import pytest
import pytz
from django.conf import settings
from fitparse.utils import FitHeaderError

from wkz.best_sections.generic import GenericBestSection
from wkz.file_helper.fit_parser import FITParser, LapData

tz = pytz.timezone(settings.TIME_ZONE)


def test__parse_metadata(fit_parser):
    p = fit_parser()
    assert p.file_name == "example.fit"


def test__parse_records(fit_parser):
    p = fit_parser()
    # test single value attributes
    assert p.sport == "running"
    assert p.distance == 5.84
    assert p.duration == datetime.timedelta(seconds=3164)
    assert p.date == datetime.datetime(2019, 9, 14, 16, 15, tzinfo=tz)
    assert p.avg_heart_rate == 130
    assert p.avg_speed == 1.845
    assert p.avg_cadence == 64
    assert p.avg_temperature == 27
    assert p.calories == 432
    assert p.aerobic_training_effect == 2.7
    assert p.anaerobic_training_effect == 0.3
    # check lengths of list attributes
    assert len(p.heart_rate_list) == 1224
    assert len(p.altitude_list) == 1224
    assert len(p.latitude_list) == 1224
    assert len(p.longitude_list) == 1224
    assert len(p.distance_list) == 1224
    assert len(p.cadence_list) == 1224
    assert len(p.temperature_list) == 1224
    assert len(p.speed_list) == 1224
    assert len(p.timestamps_list) == 1224
    # sanity check to see if element in list attributes
    assert 1.605 in p.speed_list
    assert 8.697221484035255 in p.longitude_list
    assert 49.40601873211563 in p.latitude_list
    assert 1.6 in p.distance_list
    assert 248.9 in p.altitude_list
    assert 99 in p.heart_rate_list
    assert 61 in p.cadence_list
    assert 31 in p.temperature_list
    assert 1568467325.0 in p.timestamps_list
    # check laps
    assert p.laps[0] == LapData(
        start_time=datetime.datetime(2019, 9, 14, 15, 22, 5, tzinfo=tz),
        end_time=datetime.datetime(2019, 9, 14, 15, 29, 52, tzinfo=tz),
        elapsed_time=datetime.timedelta(seconds=465, microseconds=90000),
        distance=1000.0,
        start_lat=49.405793594196446,
        start_long=8.694087238982322,
        end_long=8.696012729778888,
        end_lat=49.40549436025322,
        trigger="distance",
        speed=2.15,
    )
    # check total ascent/descent
    assert p.total_ascent == 234
    assert p.total_descent == 233


def test_set_min_max_values(fit_parser, monkeypatch):
    # mock away post processing in order to test before and after state
    def post_process(baz):
        return "foo"

    monkeypatch.setattr(FITParser, "_post_process_data", post_process)

    p = fit_parser()
    # sanity checks
    assert 61 in p.cadence_list
    assert p.avg_cadence == 64
    assert p.max_cadence is None
    assert p.min_cadence is None

    # check min max values
    p._save_data_to_dataframe()
    p._set_min_max_values()
    assert p.max_cadence == 116.0
    assert p.min_cadence == 0.0
    assert p.max_speed == 3.57
    assert p.min_speed == 0.0
    assert p.max_temperature == 31.0
    assert p.min_temperature == 26.0
    assert p.max_altitude == 353.3
    assert p.min_altitude == 238.2
    assert p.max_heart_rate == 160.0
    assert p.min_heart_rate == 95.0
    with pytest.raises(AttributeError):
        assert p.max_timestamps == 0.0
    with pytest.raises(AttributeError):
        assert p.max_distance == 0.0
    with pytest.raises(AttributeError):
        assert p.max_coordinates == 0.0


def test_drop_rows_where_all_records_are_null(fit_parser):
    p = fit_parser()
    # null rows should already be dropped, can only verify that
    # there are no more rows will all null records present
    pd.testing.assert_frame_equal(p.dataframe, p.dataframe.dropna(how="all"))


def test_convert_list_of_nones_to_empty_list(fit_parser, monkeypatch):
    # mock away post processing in order to test before and after state
    def post_process(baz):
        return "foo"

    monkeypatch.setattr(FITParser, "_post_process_data", post_process)

    p = fit_parser("with_nones.fit")
    assert p.altitude_list[:3] == [None, None, None]
    p._save_data_to_dataframe()
    assert p.dataframe.altitude_list.isna().all()
    p._convert_list_of_nones_to_empty_list()
    assert p.altitude_list == []


def test_get_fastest_sections(fit_parser):
    p = fit_parser()

    # check that fastest sections dict is empty
    assert p.best_sections == []

    p.get_best_sections()

    assert p.best_sections != []

    sec1 = GenericBestSection(1000, 622, 712, 2.89, "fastest")
    sec2 = GenericBestSection(2000, 537, 814, 2.32, "fastest")
    sec3 = GenericBestSection(3000, 428, 937, 2.13, "fastest")
    sec4 = GenericBestSection(5000, 76, 1167, 1.84, "fastest")
    sec5 = GenericBestSection(100, 555, 580, 13.05, "climb")
    sec6 = GenericBestSection(200, 535, 578, 9.73, "climb")
    sec7 = GenericBestSection(500, 469, 578, 6.75, "climb")
    sec8 = GenericBestSection(1_000, 339, 580, 5.78, "climb")
    sec9 = GenericBestSection(2_000, 50, 584, 5.0, "climb")

    assert p.best_sections == [sec1, sec2, sec3, sec4, sec5, sec6, sec7, sec8, sec9]


def test__set_avg_values(fit_parser, monkeypatch):
    # mock away post processing in order to test before and after state
    def post_process(baz):
        return "foo"

    monkeypatch.setattr(FITParser, "_post_process_data", post_process)

    p = fit_parser()

    # in test fit file average speed is present
    assert p.avg_speed == 1.845
    assert p.avg_heart_rate == 130
    assert p.avg_cadence == 64
    assert p.avg_temperature == 27

    # however, remove avg attributes to test it will be set during post processing
    p.avg_speed = None
    p.avg_heart_rate = None
    p.avg_cadence = None
    p.avg_temperature = None

    assert p.avg_speed is None
    assert p.avg_heart_rate is None
    assert p.avg_cadence is None
    assert p.avg_temperature is None

    assert not hasattr(p, "dataframe")

    p._save_data_to_dataframe()
    assert hasattr(p, "dataframe")

    p._set_avg_values()

    # now avg and min max values should be present
    assert p.avg_speed == 1.71
    assert p.avg_heart_rate == 129.72
    assert p.avg_cadence == 63.96
    assert p.avg_temperature == 26.91


def test_faulty_fit_file(tmp_path, fit_parser):
    # create a faulty fit file
    fit = tmp_path / "faulty.fit"
    content = "no valid fit file content"
    fit.write_text(content)
    assert fit.read_text() == content

    # now run fit_parser on faulty file and verify that FitHeaderError is raised
    with pytest.raises(FitHeaderError):
        fit_parser(fit)



import os
import subprocess

import pytest
from rest_framework.test import APIClient

from wkz import models
from wkz.file_helper import fit_collector


@pytest.fixture
def client():
    yield APIClient()


def test_missing_endpoint(client):
    res = client.post("/this-endpoint-is-not-implemented/")

    # having a redirection in place for all 404s will actually give as a 302 "Found (Previously "Moved temporarily")"
    assert res.status_code == 302


def test_stop(client):
    with pytest.raises(KeyboardInterrupt):
        client.post("/stop/")


def test_mount_device__failure(db, monkeypatch, client):
    # mock output of subprocess to prevent function from failing
    def dummy_output(dummy):
        return "dummy-string"

    monkeypatch.setattr(subprocess, "check_output", dummy_output)
    res = client.post("/mount-device/")

    # mounting a device is barely possible in testing, thus at least assert that the endpoint returns 500
    assert res.status_code == 500


def test_mount_device__success(db, monkeypatch, tmpdir, client):
    # prepare settings
    target_dir = tmpdir.mkdir("tracks")
    settings = models.get_settings()
    settings.path_to_garmin_device = tmpdir  # source
    settings.path_to_trace_dir = target_dir  # target
    settings.save()

    def check_output(dummy):
        return "dummy\nstring\nsome\ncontent\ncontaining\nGarmin"

    monkeypatch.setattr(subprocess, "check_output", check_output)

    # mock output of subprocess to prevent function from failing
    def try_to_mount_device():
        return "dummy-string"

    monkeypatch.setattr(fit_collector, "try_to_mount_device", try_to_mount_device)

    # mock output of actual mounting command
    def mount(bus, dev):
        return "Mounted"

    monkeypatch.setattr(fit_collector, "_mount_device_using_gio", mount)

    # create directory to import the fit files from
    fake_device_dir = os.path.join(tmpdir, "mtp:host/Primary/GARMIN/Activity/")
    os.makedirs(fake_device_dir)

    res = client.post("/mount-device/")
    assert res.status_code == 200



import json
import logging
import os
from pathlib import Path
from types import ModuleType
from typing import Dict, List, Tuple, Union

from dask.distributed import Client, as_completed
from django.db.models import Model
from fitparse.utils import FitEOFError, FitHeaderError

from wkz import configuration
from wkz.best_sections.generic import GenericBestSection
from wkz.demo import finalize_demo_activity_insertion
from wkz.file_helper.auto_naming import get_automatic_name
from wkz.file_helper.fit_parser import FITParser
from wkz.file_helper.gpx_parser import GPXParser
from wkz.file_helper.parser import Parser
from wkz.tools import sse
from wkz.tools.utils import calc_md5, limit_string, sanitize

log = logging.getLogger(__name__)


sport_naming_map = {
    "Jogging": ["jogging", "running"],
    "Cycling": ["cycle", "cycling", "biking"],
    "Mountainbiking": [
        "mountainbiking",
        "mountainbike",
        "mountain biking",
        "mountain bike",
        "mountain-biking",
        "mountain-bike",
        "mtbing",
        "mtb",
        "cycling_mountain",
    ],
    "Hiking": ["hiking", "hike", "wandern", "walking", "mountaineering"],
    "Triathlon": ["triathlon", "tria"],
    "Swimming": ["swimming", "swim", "pool"],
    "Yoga": ["yoga", "yogi"],
    "Workout": ["training"],
}


def _save_laps_to_model(lap_model, laps: list, trace_instance, update_existing: bool):
    if update_existing:
        for lap in laps:
            lap_model.objects.update_or_create(
                trace=trace_instance,
                start_time=lap.start_time,
                end_time=lap.end_time,
                elapsed_time=lap.elapsed_time,
                trigger=lap.trigger,
                start_lat=lap.start_lat,
                start_long=lap.start_long,
                end_lat=lap.end_lat,
                end_long=lap.end_long,
                distance=lap.distance,
                defaults={"speed": lap.speed},  # only speed could really be updated
            )
    else:
        for lap in laps:
            lap_object = lap_model(
                start_time=lap.start_time,
                end_time=lap.end_time,
                elapsed_time=lap.elapsed_time,
                trigger=lap.trigger,
                start_lat=lap.start_lat,
                start_long=lap.start_long,
                end_lat=lap.end_lat,
                end_long=lap.end_long,
                distance=lap.distance,
                speed=lap.speed,
                trace=trace_instance,
            )
            lap_object.save()


def _save_best_sections_to_model(best_section_model, parser, activity_instance, update_existing: bool):
    if update_existing:
        log.debug(f"updating best sections for: {activity_instance.name}")
        for section in parser.best_sections:
            best_section_model.objects.update_or_create(
                activity=activity_instance,
                kind=section.kind,
                distance=section.distance,
                defaults={
                    "start": section.start,
                    "end": section.end,
                    "max_value": section.max_value,
                },
            )
        # If the parser does not have some best sections but the db has -> delete them from the db
        db_sections = best_section_model.objects.filter(activity=activity_instance)
        for section in db_sections:
            sec = GenericBestSection(section.distance, section.start, section.end, section.max_value, section.kind)
            if sec not in parser.best_sections:
                log.debug(f"deleting section: {section} from db, because it is not present in parser")
                section.delete()
    else:
        # save best sections to model
        for section in parser.best_sections:
            best_section_object = best_section_model(
                activity=activity_instance,
                kind=section.kind,
                distance=section.distance,
                start=section.start,
                end=section.end,
                max_value=section.max_value,
            )
            best_section_object.save()


def _map_sport_name(sport_name, map_dict):  # will be adapted in GH #4
    sport = None
    for k, v in map_dict.items():
        if sanitize(sport_name) in v:
            sport = k
    if sport:
        log.debug(f"mapped activity sport: {sport_name} to {sport}")
    else:
        sport = "unknown"
        log.warning(f"could not map '{sport_name}' to given sport names, use unknown instead")
    return sport


def _save_activity_to_model(models, parser, trace_instance, importing_demo_data: bool, update_existing: bool):
    if update_existing:
        # name should not be overwritten
        activity_object = models.Activity.objects.get(trace_file=trace_instance)
        log.debug(f"updating activity attributes for: '{activity_object.name}'")
        activity_object.duration = parser.duration
        activity_object.distance = parser.distance
        if activity_object.is_demo_activity:
            log.debug(f"won't update date of demo activity: '{activity_object.name}' (ID: {activity_object.pk})")
        else:
            log.debug(f"updating date of non-demo activity: '{activity_object.name}' (ID: {activity_object.pk})")
            activity_object.date = parser.date
    else:
        # get appropriate sport from db
        sport = parser.sport
        mapped_sport = _map_sport_name(sport, sport_naming_map)
        sport_instance = models.Sport.objects.filter(slug=sanitize(mapped_sport)).first()
        if sport_instance is None:  # needs to be adapted in GH #4
            sport_instance = models.default_sport(return_pk=False)
        # determine automatic name (based on location and daytime)
        activity_object = models.Activity(
            name=get_automatic_name(parser, mapped_sport),
            sport=sport_instance,
            date=parser.date,
            duration=parser.duration,
            distance=parser.distance,
            trace_file=trace_instance,
            is_demo_activity=importing_demo_data,
        )
    activity_object.save()
    return activity_object


def _save_trace_to_model(traces_model, md5sum: str, parser, trace_file, update_existing: bool):
    parser = _convert_list_attributes_to_json(parser)
    if update_existing:
        trace_object = traces_model.objects.get(md5sum=md5sum)
        for attribute, value in parser.__dict__.items():
            if attribute == "sport":
                continue
            if hasattr(trace_object, attribute):
                db_value = getattr(trace_object, attribute)
                log.debug(
                    f"overwriting value for {attribute} old: {limit_string(db_value, 50)} "
                    f"to: {limit_string(value, 50)}"
                )
                setattr(trace_object, attribute, value)
    else:
        log.debug(f"saving trace file {trace_file} to traces model")
        trace_object = traces_model(
            path_to_file=trace_file,
            md5sum=md5sum,
            calories=parser.calories,
            # coordinates
            latitude_list=parser.latitude_list,
            longitude_list=parser.longitude_list,
            # distances
            distance_list=parser.distance_list,
            # altitude
            altitude_list=parser.altitude_list,
            max_altitude=parser.max_altitude,
            min_altitude=parser.min_altitude,
            # heart rate
            heart_rate_list=parser.heart_rate_list,
            min_heart_rate=parser.min_heart_rate,
            avg_heart_rate=parser.avg_heart_rate,
            max_heart_rate=parser.max_heart_rate,
            # cadence
            cadence_list=parser.cadence_list,
            min_cadence=parser.min_cadence,
            avg_cadence=parser.avg_cadence,
            max_cadence=parser.max_cadence,
            # speed
            speed_list=parser.speed_list,
            min_speed=parser.min_speed,
            avg_speed=parser.avg_speed,
            max_speed=parser.max_speed,
            # temperature
            temperature_list=parser.temperature_list,
            min_temperature=parser.min_temperature,
            avg_temperature=parser.avg_temperature,
            max_temperature=parser.max_temperature,
            # training effect
            aerobic_training_effect=parser.aerobic_training_effect,
            anaerobic_training_effect=parser.anaerobic_training_effect,
            # ascent/descent
            total_ascent=parser.total_ascent,
            total_descent=parser.total_descent,
            # timestamps
            timestamps_list=parser.timestamps_list,
        )
    trace_object.save()
    return trace_object


def _convert_list_attributes_to_json(parser):
    for attribute, values in parser.__dict__.items():
        if attribute.endswith("_list"):
            setattr(parser, attribute, json.dumps(values))
    return parser


def _get_md5sums_from_model(traces_model) -> List[str]:
    return list(traces_model.objects.all().values_list("md5sum", flat=True))


def _parse_data(file: Path, md5sum: str) -> Union[FITParser, GPXParser]:
    file = str(file)
    log.info(f"importing {file} ...")
    if file.lower().endswith(".gpx"):
        log.debug("parsing GPX file ...")
        parser = GPXParser(path_to_file=file, md5sum=md5sum)
    elif file.lower().endswith(".fit"):
        log.debug("parsing FIT file ...")
        parser = FITParser(path_to_file=file, md5sum=md5sum)
    else:
        log.error(f"file type: {file} unknown")
        raise NotImplementedError(
            f"Cannot parse {file} files. The only supported file formats are: {configuration.supported_formats}."
        )
    # parse best sections
    parser.get_best_sections()
    log.debug(f"finished parsing file {file}.")
    return parser


def _get_all_files(path: Path) -> List[Path]:
    trace_files = [
        Path(os.path.join(root, name))
        for root, dirs, files in os.walk(path)
        for name in files
        if name.lower().endswith(tuple(configuration.supported_formats))
    ]
    return trace_files


def _parse_single_file(
    path_to_file: Path,
    path_to_traces: Path,
    md5sum: str,
) -> Union[Parser, None]:
    """
    Parses a single file and returns the results as a Parser object. This function does
    not access the database.

    Parameters
    ----------
    path_to_file : Path
        path to the file to be parsed
    path_to_traces: Path
        path to the directory containing the activity files
    md5sum: str
        computed md5sum of the file to be parsed

    Returns
    -------
    Union[Parser, None]
        Parser object containing the payload data of the parsed file or None in case parsing fails.
    """
    try:
        parsed_data = _parse_data(str(path_to_file), md5sum)
        return parsed_data
    except (FitHeaderError, FitEOFError, AttributeError):
        # FitHeaderError and FitEOFError is used to catch corrupted fit files (e.g. non-fit files having a .fit ending).
        # AttributeError is raised in case of e.g. wahoo files, which are currently not supported by fitparse,
        # see https://github.com/dtcooper/python-fitparse/issues/121.
        sse.send(
            f"Failed to parse fit file <code>{path_to_file.relative_to(path_to_traces)}</code>. File could either be "
            f"corrupted or does not comply with the fit standard.",
            "red",
            "ERROR",
        )


def _save_single_parsed_file_to_db(
    parsed_file: Parser, models: ModuleType, importing_demo_data: bool, update_existing: bool
) -> None:
    log.debug(f"saving data of file {parsed_file.file_name} to db...")
    # save trace data to model
    trace_file_object = _save_trace_to_model(
        traces_model=models.Traces,
        md5sum=parsed_file.md5sum,
        parser=parsed_file,
        trace_file=parsed_file.path_to_file,
        update_existing=update_existing,
    )
    trace_file_instance = models.Traces.objects.get(pk=trace_file_object.pk)

    # save laps to model
    _save_laps_to_model(
        lap_model=models.Lap, laps=parsed_file.laps, trace_instance=trace_file_instance, update_existing=update_existing
    )
    # save activity itself to model
    activity_object = _save_activity_to_model(
        models=models,
        parser=parsed_file,
        trace_instance=trace_file_instance,
        importing_demo_data=importing_demo_data,
        update_existing=update_existing,
    )
    activity_instance = models.Activity.objects.get(pk=activity_object.pk)
    # save best sections to model
    _save_best_sections_to_model(
        best_section_model=models.BestSection,
        parser=parsed_file,
        activity_instance=activity_instance,
        update_existing=update_existing,
    )
    return activity_instance


def _check_and_parse_file(
    path_to_file: Path, path_to_traces: Path, md5sums_from_db: List[str], reimporting: bool
) -> Tuple[str, Path, Union[Parser, None]]:
    if reimporting:
        md5sum = calc_md5(path_to_file)
        return md5sum, path_to_file, _parse_single_file(path_to_file, path_to_traces, md5sum)
    else:
        md5sum = calc_md5(path_to_file)
        if md5sum not in md5sums_from_db:
            return md5sum, path_to_file, _parse_single_file(path_to_file, path_to_traces, md5sum)
        else:
            return md5sum, path_to_file, None


def run_importer__dask(models: ModuleType, importing_demo_data: bool = False, reimporting: bool = False) -> None:
    path_to_traces = models.get_settings().path_to_trace_dir
    log.debug(f"triggered file importer on path: {path_to_traces}")

    trace_files = _get_all_files(path_to_traces)
    _send_initial_info(len(trace_files), path_to_traces)
    md5sums_from_db = _get_md5sums_from_model(models.Traces)
    num = 0

    # check whether all files are in db already or if a single new file was added
    if _all_files_in_db_already(trace_files, md5sums_from_db) and not reimporting:
        _send_result_info(num)
        return

    seen_md5sums = {}
    if trace_files:
        with Client(processes=False, threads_per_worker=1, n_workers=1) as client:
            distributed_results = client.map(
                _check_and_parse_file,
                trace_files,
                path_to_traces=path_to_traces,
                md5sums_from_db=md5sums_from_db,
                reimporting=reimporting,
            )
            total_num = len(trace_files)
            # loop over futures in a sequential fashion to store data to sqlite db sequentially
            for i, future in enumerate(as_completed(distributed_results)):
                md5sum, path_to_file, parsed_file = future.result()
                # keep track of the seen md5sums and their file path
                seen_md5sums = _keep_track_of_md5sums_and_warn_about_duplicates(seen_md5sums, path_to_file, md5sum)
                # check if result is not None (due to failed parsing)
                if parsed_file:
                    # write parsed file to db if it does not exist yet, or in case of reimporting (=overwriting)
                    if _should_be_written_to_db(parsed_file, models.Traces, reimporting):
                        _save_single_parsed_file_to_db(parsed_file, models, importing_demo_data, reimporting)
                        log.info(f"saved activity {i+1}/{total_num} to db")
                        num += 1
                if (num + 1) % configuration.num_activities_in_progress_update == 0:
                    msg = f"<b>Progress Update:</b> Imported {num + 1} files."
                    sse.send(msg, "blue", "DEBUG")
    _send_result_info(num)

    if importing_demo_data:
        finalize_demo_activity_insertion(models)


def _all_files_in_db_already(trace_files: List[Path], md5sums_from_db: List[str]) -> bool:
    md5sums_from_files = []
    for trace in trace_files:
        md5sums_from_files.append(calc_md5(trace))
    if set(md5sums_from_db) >= set(md5sums_from_files):
        return True
    else:
        return False


def _keep_track_of_md5sums_and_warn_about_duplicates(
    seen_md5sums: Dict[str, Path], path_to_file: Path, md5sum: str
) -> Dict[str, Path]:
    if md5sum in seen_md5sums.keys():
        msg = (
            f"The following two files have the same checksum, you might want to remove one of them:<ul>"
            f"<li><code>{path_to_file}</code> and </li>"
            f"<li><code>{seen_md5sums[md5sum]}</code></li></ul>"
        )
        sse.send(msg, "yellow", "WARNING")
    else:
        seen_md5sums[md5sum] = path_to_file
    return seen_md5sums


def _should_be_written_to_db(parsed_file: Parser, traces_model: Model, reimporting: bool) -> bool:
    if reimporting:
        return True
    else:
        if traces_model.objects.filter(md5sum=parsed_file.md5sum).exists():
            existing_trace = traces_model.objects.get(md5sum=parsed_file.md5sum)
            msg = (
                f"The following two files have the same checksum, you might want to remove one of them:<ul>"
                f"<li><code>{existing_trace.path_to_file}</code> and </li>"
                f"<li><code>{parsed_file.path_to_file}</code></li></ul>"
            )
            sse.send(msg, "yellow", "WARNING")
            return False
        else:
            return True


def _send_result_info(number_of_updated: int) -> None:
    if number_of_updated == 0:
        msg = "<b>Finished File Import:</b> No new files imported."
    else:
        msg = f"<b>Finished File Import:</b> Imported {number_of_updated} new file(s)."
    sse.send(msg, "green", "DEBUG")


def _send_initial_info(number_of_activities: int, path_to_trace_dir: str):
    sse.send(
        f"<b>File Import started:</b> Found {number_of_activities} activity files in "
        f"<code>{path_to_trace_dir}</code>. Please wait.",
        "blue",
    )



import logging
import re

from django_eventstream import send_event

log = logging.getLogger(__name__)


def clean_html(raw_html):
    cleaner = re.compile("<.*?>")
    cleantext = re.sub(cleaner, "", raw_html)
    return cleantext


def send(text: str, color: str, log_level: str = "DEBUG"):
    """Server Sent Event"""
    log_level = getattr(logging, log_level)
    log.log(log_level, clean_html(text))
    send_event("event", "message", {"text": text, "color": color})



import datetime
import logging
import os
import shutil
from distutils.dir_util import copy_tree
from types import ModuleType

import pytz
from django.db import models

from workoutizer import settings as django_settings

log = logging.getLogger(__name__)


sport_data = {
    "name": ["Hiking", "Swimming", "Cycling", "Jogging"],
    "color": ["#6BD098", "#51CBCE", "#FCC468", "#F17E5D"],
    "icon": ["hiking", "swimmer", "bicycle", "running"],
    "slug": ["hiking", "swimming", "cycling", "jogging"],
    "evaluates_for_awards": [False, True, True, True],
}


def insert_demo_sports_to_model(models):
    # also insert default unknown sport
    models.default_sport()
    for i in range(len(sport_data["name"])):
        models.Sport.objects.get_or_create(
            name=sport_data.get("name")[i],
            color=sport_data.get("color")[i],
            icon=sport_data.get("icon")[i],
            slug=sport_data.get("slug")[i],
            evaluates_for_awards=sport_data.get("evaluates_for_awards")[i],
        )


def copy_demo_fit_files_to_track_dir(source_dir: str, targe_dir: str, list_of_files_to_copy: list = []):
    log.debug(f"copying demo activities from {source_dir} to {targe_dir}")
    if not list_of_files_to_copy:
        copy_tree(source_dir, targe_dir)
    else:
        for file in list_of_files_to_copy:
            shutil.copy2(os.path.join(source_dir, file), targe_dir)


def change_date_of_demo_activities(every_nth_day: int, activities: models.QuerySet):
    today = datetime.datetime.now(pytz.timezone(django_settings.TIME_ZONE))
    for i, activity in enumerate(activities.filter(is_demo_activity=True)):
        # have an activity each 3rd day
        activity.date = today - datetime.timedelta(days=i * every_nth_day + 1)
        activity.save()


def insert_custom_demo_activities(count: int, every_nth_day: int, activity_model, sport_model):
    today = datetime.datetime.now(pytz.timezone(django_settings.TIME_ZONE))
    sport = sport_model.objects.get_or_create(name="Swimming", slug="swimming", icon="swimmer", color="#51CBCE")
    for i in range(count):
        activity = activity_model(
            name="Swimming",
            date=today - datetime.timedelta(days=i * every_nth_day + 2),
            sport=sport[0],
            distance=2.0,
            duration=datetime.timedelta(minutes=90),
            is_demo_activity=True,
            description="Swimming training in my awesome pool.",
        )
        activity.save()


def prepare_import_of_demo_activities(models, list_of_files_to_copy: list = []):
    settings = models.get_settings()
    insert_demo_sports_to_model(models)
    copy_demo_fit_files_to_track_dir(
        source_dir=django_settings.INITIAL_TRACE_DATA_DIR,
        targe_dir=settings.path_to_trace_dir,
        list_of_files_to_copy=list_of_files_to_copy,
    )


def finalize_demo_activity_insertion(models: ModuleType) -> None:
    demo_activities = models.Activity.objects.filter(is_demo_activity=True)
    change_date_of_demo_activities(every_nth_day=3, activities=demo_activities)
    insert_custom_demo_activities(count=9, every_nth_day=3, activity_model=models.Activity, sport_model=models.Sport)
    log.info("finished inserting demo data")



import math

import pytest
from sportgems import DistanceTooSmallException, find_best_climb_section

from wkz import configuration
from wkz.best_sections.climb import get_best_climb_section


def test_sportgems_climb_interface__dummy_data():
    # test sportgems interface with the example data given in the repo readme
    coordinates = [(48.123, 9.35), (48.123, 9.36), (48.123, 9.37), (48.123, 9.38)]
    altitudes = [123.4, 234.5, 345.6, 456.7]
    times = [1608228953.8, 1608228954.8, 1608228955.8, 1608228956.8]

    result = find_best_climb_section(1000, times, coordinates, altitudes, tolerance=1000)

    assert result.start == 0
    assert result.end == 3
    assert math.isclose(result.climb, 4444.0, abs_tol=0.01)


def test_sportgems_climb_interface__real_activity_data__fit(fit_parser):
    parser = fit_parser()
    coordinates = list(zip(parser.latitude_list, parser.longitude_list))

    result = find_best_climb_section(1000, parser.timestamps_list, coordinates, parser.altitude_list)

    assert result.start == 339
    assert result.end == 580
    assert math.isclose(result.climb, 5.786, abs_tol=0.01)


def test_sportgems_climb_interface__real_activity_data__gpx(gpx_parser):
    parser = gpx_parser()
    coordinates = list(zip(parser.latitude_list, parser.longitude_list))

    result = find_best_climb_section(1000, parser.timestamps_list, coordinates, parser.altitude_list)

    assert result.start == 175
    assert result.end == 239
    assert math.isclose(result.climb, 10.617, abs_tol=0.01)


def test_get_best_climb_section__fit(fit_parser):
    parser = fit_parser()
    assert parser.distance == 5.84

    # test fastest 1km
    res = get_best_climb_section(1000, parser)
    assert res.start == 339
    assert res.end == 580
    assert round(res.max_value, 2) == 5.78

    # test fastest 2km
    res = get_best_climb_section(2000, parser)
    assert res.start == 50
    assert res.end == 584
    assert round(res.max_value, 2) == 5.0

    # test fastest 5km
    res = get_best_climb_section(5000, parser)
    assert res.start == 52
    assert res.end == 1151
    assert round(res.max_value, 2) == 2.36

    # test fastest 10km, in this case the activity data is shorter
    # than 10km and thus we expect that no suitable section was found
    with pytest.raises(
        DistanceTooSmallException, match="Distance of provided input data is too small for requested desired distance."
    ):
        get_best_climb_section(10_000, parser)


def test_get_best_climb_section__gpx(gpx_parser):
    parser = gpx_parser()
    assert parser.distance == 4.3

    # test fastest 1km
    res = get_best_climb_section(1000, parser)
    assert res.start == 175
    assert res.end == 239
    assert round(res.max_value, 2) == 10.62

    # test fastest 2km
    res = get_best_climb_section(2000, parser)
    assert res.start == 111
    assert res.end == 240
    assert round(res.max_value, 2) == 9.64

    # test fastest 5km
    with pytest.raises(
        DistanceTooSmallException, match="Distance of provided input data is too small for requested desired distance."
    ):
        get_best_climb_section(5000, parser)


def test_sportgems_climb_interface__fit_file_without_altitude_data(fit_parser):
    parser = fit_parser("2020-09-12-11-15-46.fit")
    # set altitude list to empty list to simulate a fit file without altitude data
    parser.altitude_list = []
    for distance in configuration.climb_distances:
        result = get_best_climb_section(distance, parser)

        # sanity check that no section causes rust panic
        assert result is None



import datetime
from pathlib import Path

import pytest
from selenium.webdriver import Firefox
from selenium.webdriver.firefox.options import Options

from workoutizer import settings as django_settings


@pytest.fixture
def webdriver():
    options = Options()
    options.headless = True
    driver = Firefox(options=options)
    driver.set_window_size(1280, 1024)
    yield driver
    driver.quit()


@pytest.fixture
def take_screenshot():
    """
    Helper function which is useful when debugging selenium browser tests.
    """

    def _take(webdriver, name="screenshot"):
        path = Path(django_settings.BASE_DIR) / "selenium_ide/debugging" / Path(f"{datetime.datetime.now()}_{name}.png")
        webdriver.save_screenshot(str(path))

    return _take



import datetime
import json
import logging
from typing import List, Union

import pytz
from django.contrib import messages
from django.db.models import Sum
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.views.generic import View

from wkz import configuration as cfg
from wkz import forms, models
from wkz.gis.geo import GeoTrace, get_list_of_coordinates
from wkz.plotting.plot_history import plot_history
from wkz.plotting.plot_pie_chart import plot_pie_chart
from wkz.plotting.plot_trend import plot_trend
from wkz.tools.colors import lines_colors
from wkz.tools.utils import cut_list_to_have_same_length
from workoutizer import __version__
from workoutizer import settings as django_settings

log = logging.getLogger(__name__)


def get_all_form_field_ids():
    """
    helper function to get all ids of input form fields to avoid keyboard navigation while entering
    text into form fields.
    """

    ids = []
    all_forms = [forms.AddSportsForm, forms.EditSettingsForm, forms.AddActivityForm, forms.EditActivityForm]
    for form in all_forms:
        ids += [f"id_{field}" for field in form.base_fields.keys()]
    return ids


class WKZView(View):
    sports = models.Sport.objects.all().order_by("name")
    form_field_ids = get_all_form_field_ids()
    context = {"sports": sports, "form_field_ids": form_field_ids}


class MapView(View):
    number_of_days = None
    days_choices = None
    settings = None

    def get(self, request, list_of_activities: list):
        self.settings = models.get_settings()
        setattr(self.settings, "trace_width", django_settings.trace_line_width)
        setattr(self.settings, "trace_opacity", django_settings.trace_line_opacity)
        self.number_of_days = self.settings.number_of_days
        self.days_choices = models.Settings.days_choices
        traces = []
        for activity in list_of_activities:
            if activity.trace_file:
                coordinates = json.dumps(
                    get_list_of_coordinates(
                        json.loads(activity.trace_file.longitude_list), json.loads(activity.trace_file.latitude_list)
                    )
                )
                sport = activity.sport.name
                if coordinates != "[]":
                    traces.append(GeoTrace(pk=activity.pk, name=activity.name, sport=sport, coordinates=coordinates))
        has_traces = True if traces else False

        if traces:
            traces, colors = cut_list_to_have_same_length(traces, lines_colors, mode="fill end", modify_only_list2=True)
            traces = zip(traces, colors)
        return {
            "traces": traces,
            "settings": self.settings,
            "days": self.number_of_days,
            "choices": self.days_choices,
            "has_traces": has_traces,
        }


class PlotView:
    number_of_days = None
    days_choices = None
    settings = None

    def get_days_config(self):
        self.settings = models.get_settings()
        self.number_of_days = self.settings.number_of_days
        self.days_choices = models.Settings.days_choices

    def get_activity_data_for_plots(self, sport_id=None):
        self.get_days_config()
        now = timezone.now()
        start_datetime = now - datetime.timedelta(days=self.number_of_days)
        if sport_id:
            activities = models.Activity.objects.filter(date__range=[start_datetime, now], sport=sport_id).order_by(
                "-date"
            )
        else:
            activities = models.Activity.objects.filter(date__range=[start_datetime, now]).order_by("-date")
        return activities


class DashboardView(View, PlotView):
    template_name = "dashboard.html"
    sports = models.Sport.objects.all().order_by("name")

    def get(self, request):
        page = 0
        settings = models.get_settings()
        self.sports = models.Sport.objects.all().order_by("name")
        activities = self.get_activity_data_for_plots()
        summary = get_summary_of_all_activities()
        context = {
            "sports": self.sports,
            "current_page": page,
            "is_last_page": False,
            "days": self.number_of_days,
            "choices": self.days_choices,
            "summary": summary,
            "page_name": "Dashboard",
            "form_field_ids": get_all_form_field_ids(),
        }
        if activities:
            script_history, div_history = plot_history(
                activities=activities, sport_model=models.Sport, number_of_days=settings.number_of_days
            )
            script_pc, div_pc = plot_pie_chart(activities=activities)
            script_trend, div_trend = plot_trend(activities=activities, sport_model=models.Sport)
            plotting_context = {
                "script_history": script_history,
                "div_history": div_history,
                "script_pc": script_pc,
                "div_pc": div_pc,
                "script_trend": script_trend,
                "div_trend": div_trend,
                "activities_selected_for_plot": True,
            }
            return render(request, self.template_name, {**context, **plotting_context})
        else:
            log.warning("no activities found...")
            context["activities_selected_for_plot"] = False
        return render(request, self.template_name, {**context})


def settings_view(request):
    sports = models.Sport.objects.all().order_by("name")
    settings = models.get_settings()
    activities = models.Activity.objects.filter(is_demo_activity=True).count()
    form = forms.EditSettingsForm(request.POST or None, instance=settings)
    return render(
        request,
        "settings/settings.html",
        {
            "sports": sports,
            "page_name": "Settings",
            "form": form,
            "settings": settings,
            "form_field_ids": get_all_form_field_ids(),
            "delete_demos": True if activities else False,
        },
    )


def settings_form(request):
    settings = models.get_settings()
    activities = models.Activity.objects.filter(is_demo_activity=True).count()
    form = forms.EditSettingsForm(request.POST or None, instance=settings)
    if request.method == "POST":
        if form.is_valid():
            log.debug(f"got valid form: {form.cleaned_data}")
            form.save()
        else:
            log.warning(f"form invalid: {form.errors}")
    return render(
        request,
        "settings/form.html",
        {
            "form": form,
            "delete_demos": True if activities else False,
        },
    )


class HelpView(WKZView):
    template_name = "lib/help.html"

    def get(self, request):
        self.context["version"] = __version__
        self.context["page_name"] = "Help"
        return render(request, template_name=self.template_name, context=self.context)


def set_number_of_days(request, number_of_days):
    settings = models.get_settings()
    settings.number_of_days = number_of_days
    log.debug(f"number of days: {number_of_days}")
    settings.save()
    if request.META.get("HTTP_REFERER"):
        return redirect(request.META.get("HTTP_REFERER"))
    else:
        return HttpResponseRedirect(reverse("home"))


def get_summary_of_all_activities(sport_slug=None):
    all_activities = models.Activity.objects.all()
    seven_days_back = (datetime.datetime.now() - datetime.timedelta(days=7)).replace(
        tzinfo=pytz.timezone(django_settings.TIME_ZONE)
    )
    if sport_slug:
        count = models.Activity.objects.filter(sport__slug=sport_slug).count()
        duration = models.Activity.objects.filter(sport__slug=sport_slug)
        total_duration = duration.aggregate(Sum("duration"))
        total_distance = models.Activity.objects.filter(sport__slug=sport_slug).aggregate(Sum("distance"))
        seven_days_trend = (
            models.Activity.objects.filter(sport__slug=sport_slug)
            .filter(date__gt=seven_days_back)
            .aggregate(Sum("duration"))
        )
    else:
        count = all_activities.count()
        total_duration = all_activities.aggregate(Sum("duration"))
        total_distance = models.Activity.objects.all().aggregate(Sum("distance"))
        seven_days_trend = models.Activity.objects.filter(date__gt=seven_days_back).aggregate(Sum("duration"))
    duration = total_duration["duration__sum"] if total_duration["duration__sum"] else datetime.timedelta(minutes=0)
    distance = int(total_distance["distance__sum"]) if total_distance["distance__sum"] else 0
    seven_days_trend = (
        seven_days_trend["duration__sum"] if seven_days_trend["duration__sum"] else datetime.timedelta(minutes=0)
    )
    return {"count": count, "duration": duration, "distance": distance, "seven_days_trend": seven_days_trend}


def custom_400_view(request, exception=None):
    messages.error(request, f"Could not find {request.path}")
    return redirect(reverse("home"))


def custom_500_view(request, exception=None):
    sports = models.Sport.objects.all().order_by("name")
    msg = f"Error while loading {request.path}"
    log.error(msg)
    template_name = "lib/500.html"
    messages.error(request, msg)
    return render(
        request,
        template_name=template_name,
        context={"url": request.path, "sports": sports, "page_name": "Error 500", "form_field_ids": []},
    )


def get_flat_list_of_pks_of_activities_in_top_awards(filter_on_sport: Union[None, str] = None) -> List[int]:
    top_award_pks = []
    if filter_on_sport:
        sport_slugs = [filter_on_sport]
    else:
        sport_slugs = [
            sport.slug for sport in models.Sport.objects.filter(evaluates_for_awards=True).exclude(name="unknown")
        ]
    for sport in sport_slugs:
        for bs in cfg.best_sections:
            for distance in bs["distances"]:
                top_awards = models.BestSection.objects.filter(
                    activity__sport__slug=sport,
                    activity__evaluates_for_awards=True,
                    kind=bs["kind"],
                    distance=distance,
                ).order_by("-max_value")[: cfg.rank_limit]
                top_award_pks += [award.activity.pk for award in top_awards]
        # also add pks of best total ascent activities
        top_ascent_awards = (
            models.Activity.objects.filter(
                sport__slug=sport,
                evaluates_for_awards=True,
            )
            .exclude(trace_file__total_ascent=None)
            .order_by("-trace_file__total_ascent")[: cfg.rank_limit]
        )
        top_award_pks += [activity.pk for activity in top_ascent_awards]
    return list(set(top_award_pks))


def get_bulk_of_rows_for_next_page(request, page: str):
    page = int(page)
    template_name = "lib/row_bulk.html"
    current_url = request.META.get("HTTP_HX_CURRENT_URL")
    sport_slug = None
    if "sport" in current_url:
        sport_slug = current_url.split("/")[-1]

    activities, is_last_page = fetch_row_data_for_page(page_nr=page, sport_slug=sport_slug)
    top_awards = get_flat_list_of_pks_of_activities_in_top_awards()

    return render(
        request,
        template_name,
        {"activities": activities, "current_page": page + 1, "is_last_page": is_last_page, "top_awards": top_awards},
    )


def fetch_row_data_for_page(page_nr: int, sport_slug=None):
    number_of_rows = cfg.number_of_rows_per_page_in_table
    start = page_nr * number_of_rows
    end = (page_nr + 1) * number_of_rows
    log.debug(f"fetching activity data for table page {page_nr}")
    is_last_page = False
    if sport_slug:
        activities = models.Activity.objects.filter(sport__slug=sport_slug).order_by("-date")
        total_nr_of_activities = activities.count()
        activities = activities[start:end]
    else:
        activities = models.Activity.objects.all().order_by("-date")
        total_nr_of_activities = activities.count()
        activities = activities[start:end]

    # indicate whether the current page is the last one
    if end >= total_nr_of_activities:
        log.debug("reached end of the table")
        is_last_page = True
    return activities, is_last_page



from wkz.best_sections.climb import get_best_climb_section
from wkz.best_sections.fastest import get_fastest_section

# supported for parsing
supported_formats = {"gpx", "fit"}

# cannot be edited nor deleted
protected_sports = {"unknown"}

# all time series attributes
time_series_attributes = {
    "latitude_list",
    "longitude_list",
    "timestamps_list",
    "distance_list",
    "heart_rate_list",
    "altitude_list",
    "cadence_list",
    "speed_list",
    "temperature_list",
}

# will create a time series plot for these list attributes
attributes_to_create_time_series_plot_for = {
    "heart_rate_list",
    "altitude_list",
    "cadence_list",
    "speed_list",
    "temperature_list",
}

min_max_attributes = {
    "heart_rate_list",
    "altitude_list",
    "cadence_list",
    "speed_list",
    "temperature_list",
}

avg_attributes = {
    "heart_rate_list",
    "cadence_list",
    "speed_list",
    "temperature_list",
}

# configuration of best sections
rank_limit = 3

# best sections config
best_sections = [
    {
        "kind": "fastest",
        "parser": get_fastest_section,
        "distances": [  # in meter
            1_000,
            2_000,
            3_000,
            5_000,
            10_000,
        ],
    },
    {
        "kind": "climb",
        "parser": get_best_climb_section,
        "distances": [  # in meter
            100,
            200,
            500,
            1_000,
            2_000,
        ],
    },
]
available_best_section = [sec["kind"] for sec in best_sections]
fastest_distances = best_sections[0]["distances"]
climb_distances = best_sections[1]["distances"]

# with respect to the table listing activities used both on dashboard and sport page
number_of_rows_per_page_in_table = 40

# how many activities should be included in one server sent event message when importing/reimporting
num_activities_in_progress_update = 5

# time to wait for an huey task to finish in seconds
huey_timeout = 30

# reduce number of data points for activity view in order speed up page load
every_nth_value = 5

# interval in minutes for periodic file import import
file_importer_interval = 1

# interval in minutes for periodic file collector
file_collector_interval = file_importer_interval



# Generated by Django 3.1.5 on 2021-01-30 12:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wkz", "0001_squashed_0014_auto_20210112_1924"),
    ]

    operations = [
        migrations.AddField(
            model_name="activity",
            name="evaluates_for_awards",
            field=models.BooleanField(default=True, verbose_name="Consider this Activity for Awards:"),
        ),
        migrations.AddField(
            model_name="sport",
            name="evaluates_for_awards",
            field=models.BooleanField(default=True, verbose_name="Consider Awards for this Sport:"),
        ),
        migrations.AlterField(
            model_name="bestsection",
            name="section_distance",
            field=models.IntegerField(),
        ),
        migrations.AlterField(
            model_name="traces",
            name="avg_temperature",
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="traces",
            name="max_temperature",
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="traces",
            name="min_temperature",
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.DeleteModel(
            name="BestSectionTopScores",
        ),
    ]



from django.urls import reverse
from selenium.webdriver.common.by import By

from wkz import configuration as cfg
from wkz import models


def test_awards_page__complete(import_demo_data, live_server, webdriver):
    """
    Check the content of the awards page by:
    1. assert existence of different html tags,...
    2. disable awards for a bunch of activities and verify that the number of trophies changed on the awards page
    3. disable awards for entire sport and verify that the number of trophies changed on the awards page
    """

    webdriver.get(live_server.url + reverse("awards"))

    # 1. assert existence of different html tags,...
    table_header = [cell.text for cell in webdriver.find_elements_by_tag_name("th")]
    assert "RANK" in table_header
    assert "DISTANCE" in table_header
    assert "DATE" in table_header
    assert "ACTIVITY" in table_header
    assert "SPEED  " in table_header

    h4 = [h4.text for h4 in webdriver.find_elements_by_tag_name("h4")]
    # note hiking activities won't show up, since they are disabled for awards in initial_data_handler
    assert "Hiking  " not in h4
    assert "Jogging  " in h4
    assert "Cycling  " in h4

    links = [a.text for a in webdriver.find_elements_by_tag_name("a")]
    assert "Noon Hiking in Aftersteg" not in links
    assert "Noon Hiking in Kornau" not in links
    assert "Noon Hiking in Bad Schandau" not in links
    assert "Noon Jogging in Mirow" in links
    assert "Noon Jogging in Heidelberg" in links
    assert "Noon Cycling in Hinterzarten" in links
    assert "Noon Cycling in Bad Schandau" in links
    assert "Early Morning Cycling in Kochel am See" in links

    table_data = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    assert "Noon Cycling in Bad Schandau" in table_data
    assert "Noon Cycling in Hinterzarten" in table_data
    assert "Early Morning Cycling in Kochel am See" in table_data
    assert "Noon Jogging in Heidelberg" in table_data
    assert "Noon Jogging in Mirow" in table_data
    assert "42.2 km/h" in table_data

    # fastest sections distances
    assert "1km" in table_data
    assert "2km" in table_data
    assert "3km" in table_data
    assert "5km" in table_data
    assert "10km" in table_data

    assert len(webdriver.find_elements_by_class_name("fa-mountain")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-tachometer-alt")) > 0

    first_num_trophies = len(webdriver.find_elements_by_class_name("fa-trophy"))
    assert first_num_trophies > 0

    # 2. disable awards for a few jogging activities and verify that the number of trophies changed on the awards page
    sport = models.Sport.objects.get(slug="jogging")
    activities = models.Activity.objects.filter(sport=sport)[:2]

    for activity in activities:
        # verify that chosen activity actually has at least one trophy
        webdriver.get(live_server.url + f"/activity/{activity.pk}")
        assert webdriver.current_url == f"{live_server.url}/activity/{activity.pk}"
        assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0

        activity.evaluates_for_awards = False
        activity.save()

    # reconnect to awards page
    webdriver.get(live_server.url + reverse("awards"))
    assert webdriver.current_url == f"{live_server.url}/awards/"

    second_num_trophies = len(webdriver.find_elements_by_class_name("fa-trophy"))
    assert second_num_trophies > 0
    # verify that we now have less awards
    assert first_num_trophies > second_num_trophies

    # 3. disable awards for entire sport and verify that the number of trophies changed on the awards page
    sport = models.Sport.objects.get(slug="cycling")
    sport.evaluates_for_awards = False
    sport.save()

    # reconnect to awards page
    webdriver.get(live_server.url + reverse("awards"))
    assert webdriver.current_url == f"{live_server.url}/awards/"

    third_num_trophies = len(webdriver.find_elements_by_class_name("fa-trophy"))
    assert third_num_trophies > 0
    # verify that we now have less awards
    assert first_num_trophies > second_num_trophies > third_num_trophies


def test_correct_activities_are_listed_on_awards_page(import_demo_data, live_server, webdriver):
    # first check activities listed in fastest awards
    webdriver.get(live_server.url + reverse("awards"))

    th = [cell.text for cell in webdriver.find_elements_by_tag_name("th")]
    assert "RANK" in th
    assert "DISTANCE" in th
    assert "DATE" in th
    assert "ACTIVITY" in th
    assert "SPEED  " in th

    fastest_top_awards = []
    for distance in cfg.fastest_distances:
        awards = models.BestSection.objects.filter(
            distance=distance,
            activity__evaluates_for_awards=True,
            kind="fastest",
        ).order_by("-max_value")[: cfg.rank_limit]
        fastest_top_awards += list(awards)

    # get table content
    td = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]

    # verify that these activities are present in the table content
    activity_names = [award.activity.name for award in fastest_top_awards]
    for name in activity_names:
        assert name in td

    max_value = [f"{round(award.max_value * 3.6, 1)} km/h" for award in fastest_top_awards]
    for value in max_value:
        assert value in td

    activity_dates = [award.activity.date.date().strftime("%b %d, %Y") for award in fastest_top_awards]
    for date in activity_dates:
        assert date in td

    # now check the same for the climb awards, first go to climb tab
    webdriver.find_element(By.LINK_TEXT, "Best Climb Section").click()

    th = [cell.text for cell in webdriver.find_elements_by_tag_name("th")]
    assert "RANK" in th
    assert "DISTANCE" in th
    assert "DATE" in th
    assert "ACTIVITY" in th
    assert "CLIMB  " in th

    climb_top_awards = []
    for distance in cfg.climb_distances:
        awards = models.BestSection.objects.filter(
            distance=distance,
            activity__evaluates_for_awards=True,
            kind="climb",
        ).order_by("-max_value")[: cfg.rank_limit]
        climb_top_awards += list(awards)

    # get table content
    td = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]

    # verify that these activities are present in the table content
    activity_names = [award.activity.name for award in climb_top_awards]
    for name in activity_names:
        assert name in td

    activity_dates = [award.activity.date.date().strftime("%b %d, %Y") for award in climb_top_awards]
    for date in activity_dates:
        assert date in td

    max_value = [f"{round(award.max_value, 2)} m/min" for award in climb_top_awards]
    for value in max_value:
        assert value in td

    # now check the same for the total ascent awards, first go to total ascent tab
    webdriver.find_element(By.LINK_TEXT, "Total Ascent").click()

    th = [cell.text for cell in webdriver.find_elements_by_tag_name("th")]
    assert "RANK" in th
    assert "DATE" in th
    assert "ACTIVITY" in th
    assert "TOTAL ASCENT  " in th

    ascent_top_awards = []
    sports = [sport for sport in models.Sport.objects.filter(evaluates_for_awards=True).exclude(name="unknown")]
    for sport in sports:
        activities = (
            models.Activity.objects.filter(
                sport=sport,
                evaluates_for_awards=True,
            )
            .exclude(trace_file__total_ascent=None)
            .order_by("-trace_file__total_ascent")[: cfg.rank_limit]
        )
        ascent_top_awards += list(activities)

    # get table content
    td = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]

    # verify that these activities are present in the table content
    activity_names = [activity.name for activity in ascent_top_awards]
    for name in activity_names:
        assert name in td

    activity_dates = [activity.date.date().strftime("%b %d, %Y") for activity in ascent_top_awards]
    for date in activity_dates:
        assert date in td

    top_total_ascent_values = [f"{activity.trace_file.total_ascent} m" for activity in ascent_top_awards]
    for value in top_total_ascent_values:
        assert value in td



import datetime
from pathlib import Path

import pytest

from wkz.file_helper.parser import Parser
from wkz.file_importer import (
    _all_files_in_db_already,
    _check_and_parse_file,
    _convert_list_attributes_to_json,
    _get_all_files,
    _map_sport_name,
    _parse_single_file,
    sport_naming_map,
)
from wkz.tools.utils import calc_md5


def test_map_sport_name():
    assert _map_sport_name("running", sport_naming_map) == "Jogging"
    assert _map_sport_name("Running", sport_naming_map) == "Jogging"
    assert _map_sport_name("swim", sport_naming_map) == "Swimming"
    assert _map_sport_name("SUP", sport_naming_map) == "unknown"


def test_convert_list_attributes_to_json(fit_parser):
    parser = fit_parser()
    assert type(parser.timestamps_list) == list
    assert type(parser.latitude_list) == list
    assert type(parser.longitude_list) == list
    parser = _convert_list_attributes_to_json(parser)
    assert type(parser.timestamps_list) == str
    assert type(parser.latitude_list) == str
    assert type(parser.longitude_list) == str


def test_get_all_files(tmpdir):
    gpx = tmpdir.mkdir("gpx").join("test.gpx")
    fit = tmpdir.mkdir("fit").join("test.fit")
    invalid = tmpdir.mkdir("txt").join("test.txt")
    for file in [gpx, fit, invalid]:
        file.write("some-content")
    assert len(_get_all_files(tmpdir)) == 2


def test__parse_single_file(demo_data_dir, fit_file):
    path = Path(demo_data_dir) / fit_file
    payload = _parse_single_file(path, demo_data_dir, "foo")
    assert isinstance(payload, Parser)

    # check core values which should have been changed
    assert payload.path_to_file == str(path)
    assert payload.file_name is not None
    assert payload.date is not None
    assert payload.md5sum == "foo"
    assert payload.sport is not None
    assert payload.duration != datetime.timedelta(minutes=0)


def test__all_files_in_db_already(demo_data_dir):
    all_files = list(Path(demo_data_dir).iterdir())
    assert _all_files_in_db_already(all_files, []) is False

    # get all md5sums of existing files
    md5sums_from_db = []
    for trace in all_files:
        md5sums_from_db.append(calc_md5(trace))

    assert _all_files_in_db_already(all_files, md5sums_from_db) is True

    # remove only one md5sum and verify that result is False
    fewer_md5sums = md5sums_from_db[:-1]
    assert _all_files_in_db_already(all_files, fewer_md5sums) is False

    # remove one file (thus db all md5sums of existing files plus one) and verify result is True
    fewer_files = all_files[:-1]
    assert _all_files_in_db_already(fewer_files, md5sums_from_db) is True


@pytest.mark.parametrize("reimporting", (False, True))
def test__check_and_parse_file(demo_data_dir, reimporting):
    trace = Path(demo_data_dir) / "2019-09-18-16-02-35.fit"

    # file md5sum is not in db
    md5sums_from_db = []
    md5sum, path_to_file, parsed_file = _check_and_parse_file(trace, demo_data_dir, md5sums_from_db, reimporting)
    assert isinstance(md5sum, str)
    assert isinstance(path_to_file, Path)
    assert isinstance(parsed_file, Parser)

    if not reimporting:
        # file md5sum is in db
        md5sums_from_db = [calc_md5(trace)]
        md5sum, path_to_file, parsed_file = _check_and_parse_file(trace, demo_data_dir, md5sums_from_db, reimporting)
        assert isinstance(md5sum, str)
        assert isinstance(path_to_file, Path)
        assert parsed_file is None



import datetime
import os

from gpxpy.gpxfield import SimpleTZ

from wkz.best_sections.generic import GenericBestSection


def test__parse_metadata(gpx_parser):
    parser = gpx_parser()
    assert parser.file_name == "example.gpx"


def test_gpx_parser__get_data(gpx_parser):
    parser = gpx_parser()
    assert parser.sport == "running"
    assert parser.duration == datetime.timedelta(seconds=1502)
    assert parser.distance == 4.3
    assert parser.date == datetime.datetime(2019, 7, 12, 17, 5, 36, tzinfo=SimpleTZ("Z"))
    assert parser.total_ascent is None
    assert parser.total_descent is None


def test_get_date_from_metadata(gpx_parser, test_data_dir):
    test_file_path = os.path.join(test_data_dir, "garmin_example.gpx")
    parser = gpx_parser(path=test_file_path)
    assert parser.date == datetime.datetime(2019, 12, 16, 8, 58, 30, tzinfo=SimpleTZ("Z"))


def test_get_sport_from_type(gpx_parser, test_data_dir):
    test_file_path = os.path.join(test_data_dir, "garmin_example.gpx")
    parser = gpx_parser(path=test_file_path)
    assert parser.sport == "running"


def test__parse_coordinates(gpx_parser):
    parser = gpx_parser()
    assert parser.longitude_list[0] == 8.687453
    assert parser.latitude_list[0] == 49.405446
    assert parser.altitude_list[0] == 128.94


def test_parse_timestamps(gpx_parser):
    parser = gpx_parser()
    assert 1562951136.0 in parser.timestamps_list


def test_get_fastest_sections(gpx_parser):
    p = gpx_parser()

    # check that fastest sections list is empty
    assert p.best_sections == []

    p.get_best_sections()

    assert p.best_sections != []

    sec1 = GenericBestSection(1000, 57, 118, 3.05, "fastest")
    sec2 = GenericBestSection(2000, 54, 171, 3.04, "fastest")
    sec3 = GenericBestSection(3000, 10, 173, 2.96, "fastest")
    sec4 = GenericBestSection(100, 115, 121, 17.98, "climb")
    sec5 = GenericBestSection(200, 8, 22, 13.55, "climb")
    sec6 = GenericBestSection(500, 62, 92, 11.95, "climb")
    sec7 = GenericBestSection(1_000, 175, 239, 10.62, "climb")
    sec8 = GenericBestSection(2_000, 111, 240, 9.64, "climb")

    assert p.best_sections == [sec1, sec2, sec3, sec4, sec5, sec6, sec7, sec8]



import datetime
import json
from typing import Union

import pandas as pd

from wkz.gis.geo import get_location_name


def _get_daytime_name(date: datetime.datetime) -> str:
    hour = date.hour
    if (hour > 4) and (hour <= 8):
        return "Early Morning"
    elif (hour > 8) and (hour <= 12):
        return "Morning"
    elif (hour > 12) and (hour <= 16):
        return "Noon"
    elif (hour > 16) and (hour <= 20):
        return "Evening"
    elif (hour > 20) and (hour <= 24):
        return "Night"
    elif hour <= 4:
        return "Late Night"


def _get_sport_name(sport_name: str) -> str:
    if sport_name == "unknown":
        return "Sport"
    else:
        return sport_name.capitalize()


def _get_coordinate_not_null(coordinates: Union[str, list]):
    try:
        if isinstance(coordinates, str):
            coordinate = pd.Series(json.loads(coordinates), dtype=float).dropna().iloc[0]
        elif isinstance(coordinates, list):
            coordinate = pd.Series(coordinates, dtype=float).dropna().iloc[0]
        else:
            raise NotImplementedError(f"Input coordinates must be either of type list or str - not {type(coordinates)}.")
    except IndexError:
        coordinate = None
    return coordinate


def get_automatic_name(parser, sport_name: str) -> str:
    lat = _get_coordinate_not_null(coordinates=parser.latitude_list)
    lon = _get_coordinate_not_null(coordinates=parser.longitude_list)
    if lat and lon:
        location = get_location_name((lat, lon))
    else:
        location = None
    daytime = _get_daytime_name(parser.date)
    sport = _get_sport_name(sport_name)
    if location:
        return f"{daytime} {sport} in {location}"
    else:
        return f"{daytime} {sport}"



import datetime
import logging
import os
from pathlib import Path

from colorfield.fields import ColorField
from django.db import models
from django.template.defaultfilters import slugify
from django.utils import timezone

from wkz.file_importer import run_importer__dask
from wkz.tools import sse
from workoutizer import settings as django_settings

log = logging.getLogger(__name__)


class Sport(models.Model):
    def __str__(self):
        return self.name

    name = models.CharField(max_length=24, unique=True, verbose_name="Sport Name")
    icon = models.CharField(max_length=24, verbose_name="Icon")
    slug = models.SlugField(max_length=100, unique=True, blank=True)
    color = ColorField(default="#42FF71", verbose_name="Color")
    evaluates_for_awards = models.BooleanField(verbose_name="Consider Sport for Awards", default=True)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        super(Sport, self).save(*args, **kwargs)


class Traces(models.Model):
    def __str__(self):
        return self.file_name

    path_to_file = models.CharField(max_length=200)
    file_name = models.CharField(max_length=100, editable=False)
    md5sum = models.CharField(max_length=32, unique=True)
    calories = models.IntegerField(null=True, blank=True)
    # coordinates
    latitude_list = models.CharField(max_length=10000000000, default="[]")
    longitude_list = models.CharField(max_length=10000000000, default="[]")
    # distance
    distance_list = models.CharField(max_length=10000000000, default="[]")
    # elevation
    altitude_list = models.CharField(max_length=10000000000, default="[]")
    max_altitude = models.FloatField(blank=True, null=True)
    min_altitude = models.FloatField(blank=True, null=True)
    # heart rate
    heart_rate_list = models.CharField(max_length=10000000000, default="[]")
    avg_heart_rate = models.IntegerField(null=True, blank=True)
    max_heart_rate = models.IntegerField(null=True, blank=True)
    min_heart_rate = models.IntegerField(null=True, blank=True)
    # cadence
    cadence_list = models.CharField(max_length=10000000000, default="[]")
    avg_cadence = models.IntegerField(null=True, blank=True)
    max_cadence = models.IntegerField(null=True, blank=True)
    min_cadence = models.IntegerField(null=True, blank=True)
    # speed
    speed_list = models.CharField(max_length=10000000000, default="[]")
    avg_speed = models.FloatField(null=True, blank=True)
    max_speed = models.FloatField(null=True, blank=True)
    min_speed = models.FloatField(null=True, blank=True)
    # temperature
    temperature_list = models.CharField(max_length=10000000000, default="[]")
    avg_temperature = models.FloatField(null=True, blank=True)
    max_temperature = models.FloatField(null=True, blank=True)
    min_temperature = models.FloatField(null=True, blank=True)
    # training effect
    aerobic_training_effect = models.FloatField(blank=True, null=True)
    anaerobic_training_effect = models.FloatField(blank=True, null=True)
    # timestamps
    timestamps_list = models.CharField(max_length=10000000000, default="[]")
    # total ascent/descent
    total_ascent = models.IntegerField(null=True, blank=True)
    total_descent = models.IntegerField(null=True, blank=True)
    # other
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.file_name = self.path_to_file.split("/")[-1]
        super(Traces, self).save()


def default_sport(return_pk: bool = True):
    sport = Sport.objects.filter(slug="unknown").first()
    if not sport:
        sport = Sport(name="unknown", color="gray", icon="question-circle", slug="unknown")
        sport.save()
    if return_pk:
        return sport.pk
    else:
        return sport


class Activity(models.Model):
    def __str__(self):
        return f"{self.name} ({self.sport})"

    name = models.CharField(max_length=200, verbose_name="Activity Name", default="unknown")
    sport = models.ForeignKey(Sport, on_delete=models.SET_DEFAULT, default=default_sport, verbose_name="Sport")
    date = models.DateTimeField(blank=False, default=timezone.now, verbose_name="Date")
    duration = models.DurationField(verbose_name="Duration", default=datetime.timedelta(minutes=30))
    distance = models.FloatField(blank=True, null=True, verbose_name="Distance", default=0)
    description = models.CharField(max_length=600, blank=True, null=True, verbose_name="Description")
    trace_file = models.ForeignKey(Traces, on_delete=models.CASCADE, blank=True, null=True)
    is_demo_activity = models.BooleanField(verbose_name="Is this a Demo Activity", default=False)
    evaluates_for_awards = models.BooleanField(verbose_name="Consider Activity for Awards", default=True)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def delete(self, *args, **kwargs):
        if self.trace_file:
            self.trace_file.delete()
            log.debug(f"deleted trace object {self.trace_file}")
            if os.path.isfile(self.trace_file.path_to_file):
                os.remove(self.trace_file.path_to_file)
                log.debug(f"deleted trace file also: {self.name}")
        super(Activity, self).delete(*args, **kwargs)
        log.debug(f"deleted activity: {self.name}")


class Lap(models.Model):
    start_time = models.DateTimeField(blank=False)
    end_time = models.DateTimeField(blank=False)
    elapsed_time = models.DurationField(blank=False)
    trigger = models.CharField(max_length=120, blank=False)
    start_lat = models.FloatField(blank=True, null=True)
    start_long = models.FloatField(blank=True, null=True)
    end_lat = models.FloatField(blank=True, null=True)
    end_long = models.FloatField(blank=True, null=True)
    distance = models.FloatField(blank=True, null=True)
    speed = models.FloatField(blank=True, null=True)
    trace = models.ForeignKey(Traces, on_delete=models.CASCADE, blank=False)
    label = models.CharField(max_length=100, blank=True, null=True, verbose_name="Label")

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class BestSection(models.Model):
    """
    Contains all best sections of all activities. Best sections could be e.g. the fastest 5km of an activity. This model
    stores the start and end of each section, which is used to render the sections in the activity view.
    """

    def __str__(self):
        return f"{self.kind} {self.distance}m: {self.max_value}"

    activity = models.ForeignKey(Activity, on_delete=models.CASCADE, blank=False)
    kind = models.CharField(max_length=120, blank=False)
    distance = models.IntegerField(blank=False)
    start = models.IntegerField(blank=False)
    end = models.IntegerField(blank=False)
    max_value = models.FloatField(blank=False)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class Settings(models.Model):
    days_choices = [
        (9999, "all"),
        (1095, "3 years"),
        (730, "2 years"),
        (365, "1 year"),
        (180, "180 days"),
        (90, "90 days"),
        (30, "30 days"),
        (10, "10 days"),
    ]

    path_to_trace_dir = models.CharField(
        max_length=120, default=django_settings.TRACKS_DIR, verbose_name="Path to Traces Directory"
    )
    path_to_garmin_device = models.CharField(
        max_length=120, default="/run/user/1000/gvfs/", verbose_name="Path to Garmin Device"
    )
    number_of_days = models.IntegerField(choices=days_choices, default=30)
    delete_files_after_import = models.BooleanField(verbose_name="Delete fit Files after Copying ", default=False)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    __original_path_to_trace_dir = None

    def __init__(self, *args, **kwargs):
        super(Settings, self).__init__(*args, **kwargs)
        self.__original_path_to_trace_dir = self.path_to_trace_dir
        self.__original_path_to_garmin_device = self.path_to_garmin_device

    def save(self, force_insert=False, force_update=False, *args, **kwargs):
        super(Settings, self).save(force_insert, force_update, *args, **kwargs)
        # whenever a path changes, check if it is a valid dir and retrigger watchdog
        if self.path_to_trace_dir != self.__original_path_to_trace_dir:
            from wkz import models

            if Path(self.path_to_trace_dir).is_dir():
                run_importer__dask(models)
            else:
                sse.send(f"<code>{self.path_to_trace_dir}</code> is not a valid path.", "red", "WARNING")
        self.__original_path_to_trace_dir = self.path_to_trace_dir

        if self.path_to_garmin_device != self.__original_path_to_garmin_device:
            from wkz import models

            if Path(self.path_to_garmin_device).is_dir():
                sse.send(f"<b>Device watchdog</b> now monitors <code>{self.path_to_garmin_device}</code>.", "green")
            else:
                sse.send(f"<code>{self.path_to_garmin_device}</code> is not a valid path.", "red", "WARNING")
        self.__original_path_to_garmin_device = self.path_to_garmin_device


def get_settings():
    return Settings.objects.get_or_create(pk=1)[0]



import logging
import os
import shutil
import subprocess
from typing import List, Union

log = logging.getLogger(__name__)


class FitCollector:
    def __init__(
        self,
        path_to_garmin_device: str,
        target_location: str,
        delete_files_after_import: bool = False,
    ):
        self.path_to_garmin_device = path_to_garmin_device
        self.delete_files_after_import = delete_files_after_import
        self.activity_dir_name = "Activity"
        self.sub_dir_for_garmin_files = "garmin"
        self.target_location = os.path.join(target_location, self.sub_dir_for_garmin_files)
        if not os.path.isdir(self.target_location):
            os.makedirs(self.target_location)

    def copy_fit_files(self):
        log.debug("looking for new activity files in garmin device")
        activity_path = _find_activity_sub_dir_in_path(
            name_of_dir=self.activity_dir_name,
            path=self.path_to_garmin_device,
            depth=4,
        )
        if activity_path:
            log.debug(f"found activity dir at: {activity_path}")
            fits = [
                os.path.join(root, name)
                for root, dirs, files in os.walk(activity_path)
                for name in files
                # Some devices genereate file named .FIT, other .fit
                if name.lower().endswith(".fit")
            ]
            if fits:
                no_file_was_copied = True
                for fit in fits:
                    file_name = str(fit.split("/")[-1])
                    target_file = os.path.join(self.target_location, file_name)
                    if not os.path.isfile(target_file):
                        shutil.copy(fit, target_file)
                        log.info(f"copied file: {target_file}")
                        no_file_was_copied = False
                        if self.delete_files_after_import:
                            os.remove(fit)
                            log.debug(f"deleted fit file from device: {file_name}")
                if no_file_was_copied:
                    log.info("No new file found.")
            else:
                log.debug(f"Could not find any activity fit files at {activity_path}")
        else:
            log.warning(f"No directory named {self.activity_dir_name} found in path {self.path_to_garmin_device}")


def _find_activity_sub_dir_in_path(name_of_dir: str, path: str, depth: int = 3) -> Union[str, bool]:
    def _get_all_subfolders(paths: List[str]) -> List[str]:
        sub_dirs = []
        for path in paths:
            sub_dirs += [f for f in os.scandir(path) if f.is_dir()]
        return sub_dirs

    paths = [path]
    for _ in range(depth):
        sub_dirs = _get_all_subfolders(paths)
        for sub_dir in sub_dirs:
            if name_of_dir.lower() == sub_dir.name.lower():
                return sub_dir.path
        paths = sub_dirs
    # if this line is reach it means the dir was not found
    return False


def try_to_mount_device():
    log.debug("trying to mount device...")
    lsusb_output = subprocess.check_output("lsusb")
    split = str(lsusb_output).split("\\n")
    mount_output = None
    for line in split:
        if "Garmin" in line:
            log.debug(f"found Garmin device in: {line}")
            bus_start = line.find("Bus") + 4
            bus = line[bus_start : bus_start + 3]
            device_start = line.find("Device") + 7
            dev = line[device_start : device_start + 3]
            try:
                mount_output = _mount_device_using_gio(bus, dev)
            except subprocess.CalledProcessError as e:
                log.warning(f"could not mount device: {e}", exc_info=True)
                return None
        else:
            # log.debug(f"no Garmin device found in 'lsusb' line: {line}")
            pass
    if mount_output:
        if "Mounted" in mount_output:
            path_start = mount_output.find("at")
            mount_path = mount_output[path_start + 2 : -1]
            log.info(f"successfully mounted device at: {mount_path}")
            return mount_path
        else:
            log.warning("could not mount device")
    else:
        log.warning("Found Garmin device, but could not mount it.")


def _mount_device_using_gio(bus: str, dev: str) -> str:
    return subprocess.check_output(["gio", "mount", "-d", f"/dev/bus/usb/{bus}/{dev}"]).decode("utf-8")



import datetime

import pytest
import pytz
from django.urls import reverse
from django.utils import timezone
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from wkz import models


def test_activity_page__complete__fit(import_one_activity, live_server, webdriver):
    import_one_activity("cycling_bad_schandau.fit")

    activity = models.Activity.objects.get()
    pk = activity.pk
    webdriver.get(live_server.url + f"/activity/{pk}")

    table_header = [cell.text for cell in webdriver.find_elements_by_tag_name("th")]
    assert "#" in table_header
    assert "TIME" in table_header
    assert "DISTANCE" in table_header
    assert "PACE" in table_header
    assert "LABEL" in table_header

    # check summary facts
    card_category = [cell.text for cell in webdriver.find_elements_by_class_name("card-category")]
    assert "Date" in card_category
    assert "Distance" in card_category
    assert "Duration" in card_category
    assert "Calories" in card_category
    card_title = [cell.text for cell in webdriver.find_elements_by_class_name("card-title")]
    assert "4h 59m" in card_title
    assert "44.96 km" in card_title
    assert "1044 kcal" in card_title
    assert "29. Aug 20" in card_title

    table_data = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    # best sections
    assert "  1km" in table_data
    assert "  2km" in table_data
    assert "  3km" in table_data
    assert "  5km" in table_data
    assert "  10km" in table_data
    assert "29.7 km/h" in table_data
    assert "25.0 km/h" in table_data

    # avg + min + max values
    assert "max" in table_data
    assert "high" in table_data
    assert "avg" in table_data
    assert "min" in table_data
    assert "9.0 km/h" in table_data
    assert "47.9 km/h" in table_data
    assert "06:39 min/km" in table_data
    assert "01:15 min/km" in table_data
    assert "23.0 °C" in table_data
    assert "26.0 °C" in table_data
    assert "32.0 °C" in table_data
    assert "1" in table_data
    assert "1:46:15" in table_data
    assert "11423" in table_data
    assert "09:18" in table_data
    assert "447 m" in table_data  # total ascent
    assert "451 m" in table_data  # total descent

    assert webdriver.find_element_by_class_name("navbar-brand").text == "Noon Cycling In Bad Schandau"

    # check card titles
    assert "Fastest Sections  " in card_title
    assert "Best Climb Sections  " in card_title
    assert "Speed" in card_title
    assert "Pace" in card_title
    assert "Temperature" in card_title
    assert "Laps" in card_title
    assert "Elevation" in card_title

    links = [a.text for a in webdriver.find_elements_by_tag_name("a")]
    assert "WORKOUTIZER" in links
    assert "DASHBOARD" in links
    assert "AWARDS" in links
    assert "SPORTS" in links
    assert "ADD SPORT" in links
    assert "+" in links
    assert "−" in links
    assert "Leaflet" in links
    assert "OpenStreetMap" in links

    spans = [a.text for a in webdriver.find_elements_by_tag_name("span")]
    assert "Streets" in spans
    assert "Topo" in spans
    assert "Terrain" in spans
    assert "Satellite" in spans

    # check that icons are present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-fire")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-road")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-history")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-history")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-calendar-alt")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-arrow-down")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-arrow-up")) > 0

    # check that map is displayed
    assert (
        webdriver.find_element_by_id("leaflet_map").text
        == "Streets\nTopo\nTerrain\nSatellite\n+\n−\n3 km\nLeaflet | Map data: © OpenStreetMap"
    )

    # check that bokeh plots are available
    assert webdriver.find_element_by_class_name("bk-root").text == "Show Laps"
    assert len(webdriver.find_elements_by_class_name("bk-canvas")) == 3


def test_activity_page__complete__gpx(import_one_activity, live_server, webdriver):
    import_one_activity("cycling_walchensee.gpx")

    activity = models.Activity.objects.get()
    pk = activity.pk
    webdriver.get(live_server.url + f"/activity/{pk}")

    # check summary facts
    card_category = [cell.text for cell in webdriver.find_elements_by_class_name("card-category")]
    assert "Date" in card_category
    assert "Distance" in card_category
    assert "Duration" in card_category
    assert "Calories" in card_category
    card_title = [cell.text for cell in webdriver.find_elements_by_class_name("card-title")]
    assert "3h 32m" in card_title
    assert "82.6 km" in card_title
    assert "- kcal" in card_title
    assert "28. May 21" in card_title

    table_data = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    # best sections
    assert "  1km" in table_data
    assert "  2km" in table_data
    assert "  3km" in table_data
    assert "  5km" in table_data
    assert "  10km" in table_data
    assert "51.3 km/h" in table_data
    assert "49.9 km/h" in table_data

    assert webdriver.find_element_by_class_name("navbar-brand").text == "Early Morning Cycling In Kochel Am See"

    # check card titles
    assert "Fastest Sections  " in card_title
    assert "Best Climb Sections  " in card_title

    links = [a.text for a in webdriver.find_elements_by_tag_name("a")]
    assert "WORKOUTIZER" in links
    assert "DASHBOARD" in links
    assert "AWARDS" in links
    assert "SPORTS" in links
    assert "ADD SPORT" in links
    assert "+" in links
    assert "−" in links
    assert "Leaflet" in links
    assert "OpenStreetMap" in links

    spans = [a.text for a in webdriver.find_elements_by_tag_name("span")]
    assert "Streets" in spans
    assert "Topo" in spans
    assert "Terrain" in spans
    assert "Satellite" in spans

    # check that icons are present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-fire")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-road")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-history")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-history")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-calendar-alt")) > 0

    # check that map is displayed
    assert (
        webdriver.find_element_by_id("leaflet_map").text
        == "Streets\nTopo\nTerrain\nSatellite\n+\n−\n10 km\nLeaflet | Map data: © OpenStreetMap"
    )

    # check that only one bokeh plot is available
    assert len(webdriver.find_elements_by_class_name("bk-canvas")) == 1


def test_edit_activity_page(import_one_activity, live_server, webdriver, insert_sport):
    # add some sports
    insert_sport("Cycling")
    insert_sport("MTB")
    assert models.Sport.objects.count() == 2

    # activity will be mapped to sport "Cycling"
    import_one_activity("cycling_bad_schandau.fit")

    activity = models.Activity.objects.get()
    pk = activity.pk
    assert activity.sport.name == "Cycling"

    # save initial date
    initial_date = activity.date

    webdriver.get(live_server.url + f"/activity/{pk}")

    # verify that activity has some awards
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0
    # because it currently evaluates for awards there is no exclamation circle icon
    assert len(webdriver.find_elements_by_class_name("fa-exclamation-circle")) == 0

    # go to edit activity page by clicking the edit button
    button = webdriver.find_element_by_id("edit-activity-button")
    button.click()

    # verify url got changed to activity view
    assert webdriver.current_url == f"{live_server.url}/activity/{pk}/edit/"

    assert activity.name == "Noon Cycling in Bad Schandau"
    assert activity.duration == datetime.timedelta(seconds=17970)
    assert activity.is_demo_activity is False
    assert activity.evaluates_for_awards is True
    # make activity a demo activity to verify it won't get changed back
    activity.is_demo_activity = True
    activity.save()
    assert activity.is_demo_activity is True

    assert webdriver.find_element_by_class_name("navbar-brand").text == "Edit Activity: Noon Cycling In Bad Schandau"
    assert webdriver.find_element_by_id("submit-button").text == "  SAVE"
    assert (
        webdriver.find_element_by_tag_name("form").text
        == """Activity Name
Sport
---------
Cycling
MTB
Date
Duration [HH:MM:SS]
Distance [in km]
Description
Consider Activity for Awards

Lap Data


  SAVE
CANCEL
  DELETE"""
    )

    links = [link.text for link in webdriver.find_elements_by_tag_name("a")]
    assert "DASHBOARD" in links
    assert "ADD SPORT" in links
    assert "SPORTS" in links
    assert "AWARDS" in links
    assert "WORKOUTIZER" in links
    assert "Lap Data" in links  # means activity has some laps
    assert "CANCEL" in links
    assert "  DELETE" in links
    assert "REPORT AN ISSUE" in links
    assert "" in links

    # uncheck the box for evaluates_for_awards
    webdriver.find_element_by_class_name("form-check-label").click()

    # enter a different name
    name_field = webdriver.find_element_by_css_selector("#id_name")
    name_field.clear()
    name_field.send_keys("Dummy Activity Name")

    # change duration
    duration_field = webdriver.find_element_by_css_selector("#id_duration")
    duration_field.clear()
    duration_field.send_keys("01:11:11")

    # change sport to verify dropdown also works
    dropdown = webdriver.find_element(By.ID, "id_sport")
    dropdown.find_element(By.XPATH, "//option[. = 'MTB']").click()

    # modify the date of the activity to verify the datetime picker widget works
    webdriver.find_element(By.ID, "id_date").click()
    webdriver.find_element(By.CSS_SELECTOR, "tr:nth-child(4) > .day:nth-child(5)").click()
    webdriver.find_element(By.CSS_SELECTOR, "tr:nth-child(3) > .day:nth-child(3)").click()
    webdriver.find_element(By.CSS_SELECTOR, ".fa-clock-o").click()
    webdriver.find_element(By.CSS_SELECTOR, "td:nth-child(3) .fa-chevron-down").click()

    # also edit lap data
    webdriver.find_element(By.ID, "edit-lap-data").click()
    lap_input_0 = webdriver.find_element(By.ID, "id_form-0-label")
    lap_input_0.clear()
    lap_input_0.send_keys("lap label 0")

    # verify that there is no second element available
    with pytest.raises(NoSuchElementException, match=r'Message: Unable to locate element: |[id="id_form\-1-label"]'):
        webdriver.find_element(By.ID, "id_form-1-label")

    # submit form with modified data
    button = webdriver.find_element_by_id("submit-button")
    button.click()

    # verify url got changed to activity view
    assert webdriver.current_url == f"{live_server.url}/activity/{pk}"

    # check that all trophies got removed, since activity no longer evaluates for awards
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 0
    # two warning exclamation circles are present, one for fastest, one for climb
    assert len(webdriver.find_elements_by_class_name("fa-exclamation-circle")) == 2

    # verify attributes got changed
    activity = models.Activity.objects.get()
    assert activity.name == "Dummy Activity Name"
    assert activity.duration == datetime.timedelta(seconds=4271)
    assert activity.is_demo_activity is True
    assert activity.evaluates_for_awards is False
    assert activity.sport.name == "MTB"
    assert activity.date != initial_date

    # verify that lap label got changed
    laps = models.Lap.objects.filter(trace=activity.trace_file, trigger="manual")
    assert laps[0].label == "lap label 0"


def test_add_activity_page(insert_sport, webdriver, live_server):
    # insert some sports
    insert_sport("Cycling")
    insert_sport("MTB")
    assert models.Sport.objects.count() == 2

    webdriver.get(live_server.url)
    webdriver.find_element(By.ID, "add-activity-button").click()
    assert webdriver.current_url == live_server.url + reverse("add-activity")

    # set name
    webdriver.find_element(By.ID, "id_name").click()
    webdriver.find_element(By.ID, "id_name").clear()
    webdriver.find_element(By.ID, "id_name").send_keys("Dummy Activity")
    webdriver.find_element(By.CSS_SELECTOR, "form").click()
    # set sport
    dropdown = webdriver.find_element(By.ID, "id_sport")
    dropdown.find_element(By.XPATH, "//option[. = 'Cycling']").click()
    # set date
    webdriver.find_element(By.ID, "id_date").click()
    webdriver.find_element(By.CSS_SELECTOR, ".fa-clock-o").click()
    # change hours
    webdriver.find_element(By.CSS_SELECTOR, "td:nth-child(1) .fa-chevron-up").click()
    # change minutes
    webdriver.find_element(By.CSS_SELECTOR, "td:nth-child(3) .fa-chevron-down").click()
    # set duration
    webdriver.find_element(By.ID, "id_duration").clear()
    webdriver.find_element(By.ID, "id_duration").send_keys("00:31:00")
    # set distance
    webdriver.find_element(By.ID, "id_distance").send_keys("2.3")
    # set description
    webdriver.find_element(By.ID, "id_description").send_keys("super sport")
    webdriver.find_element(By.CSS_SELECTOR, "form").click()
    # submit and wait
    webdriver.find_element(By.ID, "submit-button").click()
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.CSS_SELECTOR, ".alert")))

    assert webdriver.current_url == live_server.url + reverse("home")
    assert "Successfully added 'Dummy Activity'" in webdriver.find_element(By.CSS_SELECTOR, ".alert").text

    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity.name == "Dummy Activity"
    assert activity.duration == datetime.timedelta(minutes=31)
    assert activity.distance == 2.3
    assert activity.description == "super sport"
    assert activity.sport.name == "Cycling"
    assert activity.date.date() == timezone.now().date()


def test_activity_page__rendering_of_sport_icon_on_map(insert_sport, import_one_activity, live_server, webdriver):
    # Test to fix issue GH72
    icon_name = "running"
    insert_sport(name="Jogging", icon=icon_name)

    import_one_activity("2019-09-18-16-02-35.fit")
    activity = models.Activity.objects.get()
    pk = activity.pk

    # check out laps of activity
    assert models.Lap.objects.count() == 2
    laps = models.Lap.objects.all()
    for lap in laps:
        # each lap has at least one coordinate
        assert lap.end_lat is not None or lap.start_lat is not None
        assert lap.end_long is not None or lap.start_long is not None
        assert lap.trigger != "manual"

    # insert an additional lap without any coordinate data
    lap_obj = models.Lap(
        start_time=activity.date,
        end_time=activity.date + datetime.timedelta(minutes=1),
        elapsed_time=-datetime.timedelta(minutes=100),
        trigger="manual",
        start_lat=None,
        start_long=None,
        end_lat=None,
        end_long=None,
        distance=1.0,
        speed=10.0,
        trace=activity.trace_file,
    )
    lap_obj.save()
    assert models.Lap.objects.count() == 3
    new_lap = models.Lap.objects.get(trigger="manual")
    assert new_lap.start_lat is None
    assert new_lap.start_long is None
    assert new_lap.end_lat is None
    assert new_lap.end_long is None

    webdriver.get(live_server.url + f"/activity/{pk}")
    assert webdriver.current_url == f"{live_server.url}/activity/{pk}"

    initial_number_of_sport_icons = len(webdriver.find_elements_by_class_name(f"fa-{icon_name}"))

    headings = [h.text for h in webdriver.find_elements_by_tag_name("h4")]
    assert "Fastest Sections  " in headings
    assert "Best Climb Sections  " in headings
    assert "Cadence" in headings
    assert "Speed" in headings
    assert "Pace" in headings
    assert "Temperature" in headings
    assert "Laps" in headings

    # hover over ploy in order to trigger the rendering of the sport icon on the leaflet map
    altitude_plot = webdriver.find_element(By.CSS_SELECTOR, ".bk:nth-child(1) > .bk > .bk-canvas-events")
    altitude_plot.click()
    hover = ActionChains(webdriver).move_to_element(altitude_plot)
    hover.perform()

    # once sport icon gets rendered we should find one more sport icons than before
    number_of_sport_icons_when_hovering = len(webdriver.find_elements_by_class_name(f"fa-{icon_name}"))
    assert number_of_sport_icons_when_hovering == initial_number_of_sport_icons + 1


def test_activity_page__missing_attributes(import_one_activity, live_server, webdriver):
    import_one_activity("cycling_bad_schandau.fit")

    activity = models.Activity.objects.get()

    # set some attributes to None
    activity.trace_file.avg_speed = None
    activity.trace_file.max_speed = None
    activity.trace_file.max_avg_cadence = None
    activity.trace_file.aerobic_training_effect = None
    activity.trace_file.avg_heart_rate = None
    activity.trace_file.min_temperature = None
    activity.trace_file.save()

    pk = activity.pk
    webdriver.get(live_server.url + f"/activity/{pk}")

    # verify that page load does not fail (without safety measures in place this would fail with TypeError, see GH95)
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Noon Cycling In Bad Schandau"

    # also verify that the sections with missing data are not displayed
    headings = [h.text for h in webdriver.find_elements_by_tag_name("h4")]
    assert "Trainings Effect" not in headings
    assert "Heart Rate" not in headings
    assert "Speed" not in headings
    assert "Cadence" not in headings
    assert "Pace" not in headings
    assert "Temperature" not in headings

    # these should still be there
    assert "Fastest Sections  " in headings
    assert "Laps" in headings

    # again add values for missing attributes
    activity = models.Activity.objects.get()
    activity.trace_file.avg_speed = 1.1
    activity.trace_file.max_speed = 2.2
    activity.trace_file.max_cadence = 3.3
    activity.trace_file.avg_cadence = 3.3
    activity.trace_file.aerobic_training_effect = 4.4
    activity.trace_file.anaerobic_training_effect = 4.4
    activity.trace_file.avg_heart_rate = 5.5
    activity.trace_file.min_heart_rate = 5.5
    activity.trace_file.max_heart_rate = 5.5
    activity.trace_file.min_temperature = 6.6
    activity.trace_file.save()

    # again open activity page
    webdriver.refresh()

    # and verify all headings are back
    headings = [h.text for h in webdriver.find_elements_by_tag_name("h4")]
    assert "Fastest Sections  " in headings
    assert "Laps" in headings
    assert "Speed" in headings
    assert "Pace" in headings
    assert "Temperature" in headings
    assert "Trainings Effect" in headings
    assert "Heart Rate" in headings
    assert "Cadence" in headings


def test_trophy_icon_for_awarded_activity_is_displayed_correctly(db, live_server, webdriver):
    assert models.Activity.objects.count() == 0

    # add activity
    sport = models.Sport.objects.create(name="Swimming", slug="swimming", icon="swimmer", color="#51CBCE")
    activity = models.Activity.objects.create(
        name="Foo",
        sport=sport,
        date=datetime.datetime(2020, 1, 1, 15, 30, 0, tzinfo=pytz.UTC),
        duration=datetime.timedelta(0, 3600),
    )

    activity_url = live_server.url + f"/activity/{activity.pk}"
    webdriver.get(activity_url)
    # neither best sections nor total ascent is available, thus we expect 0 trophies on the activity page
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 0

    # add total ascent to activity
    trace = models.Traces.objects.create(
        path_to_file="dummy/path",
        file_name="foo.baa",
        md5sum="1a2b3c4d5e",
        total_ascent=500,
        total_descent=501,
    )
    activity.trace_file = trace
    activity.save()

    webdriver.get(activity_url)
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 1

    trophy = webdriver.find_element_by_class_name("fa-trophy")
    gold = "rgb(255, 215, 0)"
    assert trophy.value_of_css_property("color") == gold

    # add best sections
    models.BestSection.objects.create(
        activity=activity,
        kind="fastest",
        distance=1000,
        start=42,
        end=128,
        max_value=123.45,
    )
    webdriver.get(activity_url)
    # since the title of the "fastest section" card will also contain an additional trophy, we now expect 3
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 3



import datetime

import numpy as np
import pytest
import pytz
from django.urls import reverse

from wkz import models
from wkz.views import (
    get_flat_list_of_pks_of_activities_in_top_awards,
    get_summary_of_all_activities,
)
from workoutizer import settings as django_settings


def test_help_view(db, client):
    response = client.get(reverse("help"))
    assert response.status_code == 200
    assert "Keyboard Navigation" in response.content.decode("UTF-8")


def test_dashboard_view(db, client):
    response = client.get(reverse("home"))
    assert response.status_code == 200
    assert "Add Activity" in response.content.decode("UTF-8")


def test_settings_view(db, client):
    response = client.get(reverse("settings"))
    assert response.status_code == 200
    assert "Settings" in response.content.decode("UTF-8")


def test_best_sections_view(db, client):
    response = client.get(reverse("awards"))
    assert response.status_code == 200
    assert "awards" in response.content.decode("UTF-8")


def test_activity_view__activity_present(db, client, settings, sport, activity):
    pk = models.Activity.objects.get().pk
    response = client.get(f"/activity/{pk}")
    assert response.status_code == 200


def test_activity_view__no_activity(db, client):
    with pytest.raises(models.Activity.DoesNotExist):
        client.get("/activity/1")


def test_get_flat_list_of_pks_of_activities_in_top_awards(db, import_demo_data):
    result_pks = get_flat_list_of_pks_of_activities_in_top_awards()
    assert len(result_pks) == 7
    assert len(np.unique(result_pks)) == 7


def test_get_summary_of_all_activities(insert_sport, db, insert_activity):
    today = datetime.datetime.now().replace(tzinfo=pytz.timezone(django_settings.TIME_ZONE))

    # without any activity
    expected = {
        "count": 0,
        "duration": datetime.timedelta(seconds=0),
        "distance": 0,
        "seven_days_trend": datetime.timedelta(seconds=0),
    }
    result = get_summary_of_all_activities()
    assert expected == result

    # insert one activity
    insert_activity(date=today - datetime.timedelta(days=3))
    expected = {
        "count": 1,
        "duration": datetime.timedelta(seconds=1800),
        "distance": 5,
        "seven_days_trend": datetime.timedelta(seconds=1800),
    }
    result = get_summary_of_all_activities()
    assert expected == result

    # insert another activity with duration != 1800 sec
    insert_activity(date=today - datetime.timedelta(days=6), duration=datetime.timedelta(hours=7))
    expected = {
        "count": 2,
        "duration": datetime.timedelta(seconds=27000),
        "distance": 10,
        "seven_days_trend": datetime.timedelta(seconds=27000),
    }
    result = get_summary_of_all_activities()
    assert expected == result

    # insert another activity which is 8 days old and thus excluded from the seven days trend
    insert_activity(date=today - datetime.timedelta(days=8), duration=datetime.timedelta(hours=1))
    expected = {
        "count": 3,
        "duration": datetime.timedelta(seconds=30600),
        "distance": 15,
        "seven_days_trend": datetime.timedelta(seconds=27000),  # stays the same
    }
    result = get_summary_of_all_activities()
    assert expected == result

    # insert activity of other sport
    sport = insert_sport(name="Jumping", icon="cat")
    insert_activity(date=today - datetime.timedelta(days=2), duration=datetime.timedelta(hours=0.1), sport=sport)
    # get summary for the original sport 'cycling' which should bring the same results
    result = get_summary_of_all_activities("cycling")
    assert expected == result

    # without filtering for the sport, we expect:
    expected = {
        "count": 4,
        "duration": datetime.timedelta(seconds=30960),
        "distance": 20,
        "seven_days_trend": datetime.timedelta(seconds=27360),
    }
    result = get_summary_of_all_activities()
    assert expected == result



import datetime

import pytest

from wkz.file_helper.auto_naming import (
    _get_coordinate_not_null,
    _get_daytime_name,
    _get_sport_name,
)


def test__get_daytime_name():
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 0)) == "Late Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 1)) == "Late Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 2)) == "Late Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 3)) == "Late Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 4)) == "Late Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 5)) == "Early Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 6)) == "Early Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 7)) == "Early Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 8)) == "Early Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 9)) == "Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 10)) == "Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 11)) == "Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 12)) == "Morning"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 13)) == "Noon"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 14)) == "Noon"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 15)) == "Noon"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 16)) == "Noon"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 17)) == "Evening"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 18)) == "Evening"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 19)) == "Evening"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 20)) == "Evening"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 21)) == "Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 22)) == "Night"
    assert _get_daytime_name(datetime.datetime(2020, 1, 1, 23)) == "Night"

    with pytest.raises(ValueError, match="hour must be in 0..23"):
        assert _get_daytime_name(datetime.datetime(2020, 1, 1, 24)) == "Late Night"


def test__get_sport_name():
    assert _get_sport_name("running") == "Running"
    assert _get_sport_name("unknown") == "Sport"
    assert _get_sport_name("KAYAKING") == "Kayaking"
    assert _get_sport_name("----") == "----"


def test__get_coordinate_not_null():
    assert _get_coordinate_not_null("[null, null, 48.123, null, null, 48.234]") == 48.123
    assert _get_coordinate_not_null("[null, null]") is None



from huey import crontab
from huey.contrib.djhuey import periodic_task

from wkz.watchdogs import trigger_device_watchdog, trigger_file_watchdog
from wkz import configuration as cfg


@periodic_task(crontab(minute=f"*/{cfg.file_importer_interval}"))
def check_for_mounted_device():
    trigger_device_watchdog()


@periodic_task(crontab(minute=f"*/{cfg.file_collector_interval}"))
def check_for_new_activity_files():
    trigger_file_watchdog()



import json


def migrate_altitudes(traces_model):
    print(f"migrating trace models altitude: {traces_model}")
    traces = traces_model.objects.all()
    for trace in traces:
        t = traces_model.objects.get(pk=trace.pk)
        if t.elevation:
            print(f"modifying trace: {t}")
            ele = json.loads(t.elevation)
            print(f"ele: {ele}")
            ele = list(ele)
            print(f"len ele: {len(ele)}")
            print(f"max ele: {min(ele)}")
            print(f"max ele: {max(ele)}")
            t.save()



from sportgems import find_best_climb_section

from wkz.best_sections.generic import GenericBestSection


def get_best_climb_section(distance: int, parser) -> GenericBestSection:
    coordinates = list(zip(parser.latitude_list, parser.longitude_list))

    # safety check to catch activities with empty altitude list
    if not parser.altitude_list:
        return None

    # call to rust
    section = find_best_climb_section(
        desired_distance=distance,
        times=parser.timestamps_list,
        coordinates=coordinates,
        altitudes=parser.altitude_list,
    )

    return GenericBestSection(
        start=section.start,
        end=section.end,
        distance=distance,
        max_value=round(section.climb, 2),
        kind="climb",
    )



import datetime

import pytest
import pytz
from django.urls import reverse
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from wkz import configuration, models
from wkz.views import get_flat_list_of_pks_of_activities_in_top_awards


def test_dashboard_page_accessible(live_server, webdriver):
    webdriver.get(live_server.url + reverse("home"))

    # first time running workoutizer will lead to the dashboard page with no data
    h3 = webdriver.find_element_by_css_selector("h3")
    assert h3.text == "No activity data selected for plotting  "


def test_add_activity_button(live_server, webdriver):
    webdriver.get(live_server.url + reverse("home"))

    # ensure button to create new data is actually redirecting to add activity page
    webdriver.find_element_by_id("add-activity-button").click()
    assert webdriver.current_url == live_server.url + reverse("add-activity")


def test_nav_bar_items(live_server, webdriver):
    webdriver.get(live_server.url + reverse("home"))

    # ensure nav bar link to settings page works
    webdriver.find_element_by_id("settings-button").click()
    assert webdriver.current_url == live_server.url + reverse("settings")

    # ensure nav bar link to help page works
    webdriver.find_element_by_id("help-button").click()
    assert webdriver.current_url == live_server.url + reverse("help")


def test_drop_down_visible(live_server, webdriver, settings):
    webdriver.get(live_server.url + reverse("home"))
    days = settings.number_of_days

    dropdown_button = webdriver.find_element_by_id("dropdown-btn")
    assert dropdown_button.text == f"{days} DAYS"


def test_dashboard_page__complete(import_demo_data, live_server, webdriver):
    webdriver.get(live_server.url + reverse("home"))
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Dashboard"

    # check that all activity names are in the table
    table_data = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    assert "Noon Jogging in Heidelberg" in table_data
    assert "Swimming" in table_data
    assert "Noon Cycling in Bad Schandau" in table_data
    assert "Noon Cycling in Hinterzarten" in table_data
    assert "Early Morning Cycling in Kochel am See" in table_data
    assert "Noon Hiking in Aftersteg" in table_data
    assert "Noon Hiking in Kornau" in table_data

    # check that the trophy icons are present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0

    # check that sport icons are present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-bicycle")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-hiking")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-running")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-swimmer")) > 0

    # check left side bar icons are present
    assert len(webdriver.find_elements_by_class_name("fa-chart-line")) == 2  # sidebar and summary facts
    assert len(webdriver.find_elements_by_class_name("fa-medal")) == 1
    assert len(webdriver.find_elements_by_class_name("fa-plus")) >= 1

    # check icons in top navbar are present
    assert len(webdriver.find_elements_by_class_name("fa-question-circle")) == 1
    assert len(webdriver.find_elements_by_class_name("fa-cog")) == 1

    # check icons in right sidebar are present
    assert len(webdriver.find_elements_by_class_name("fa-hashtag")) == 1
    assert len(webdriver.find_elements_by_class_name("fa-road")) == 1
    assert len(webdriver.find_elements_by_class_name("fa-history")) == 3

    # verify that each activity which is in top score has an award and also is displayed with a trophy
    href = [a.get_attribute("href") for a in webdriver.find_elements_by_tag_name("a")]
    pks = []
    for url in href:
        if url is not None:
            if "/activity/" in url:
                pks.append(url.split("/")[-1])
    pks = set(pks)

    top_award_pks = []
    sport_slugs = [
        sport.slug for sport in models.Sport.objects.filter(evaluates_for_awards=True).exclude(name="unknown")
    ]

    for sport in sport_slugs:
        for bs in configuration.best_sections:
            for distance in bs["distances"]:
                top_awards = models.BestSection.objects.filter(
                    activity__sport__slug=sport,
                    activity__evaluates_for_awards=True,
                    kind=bs["kind"],
                    distance=distance,
                ).order_by("-max_value")[: configuration.rank_limit]
                top_award_pks += [str(award.activity.pk) for award in top_awards]
    top_award_pks = list(set(top_award_pks))

    expected_num_of_trophies = 0
    for pk in pks:
        if pk in top_award_pks:
            expected_num_of_trophies += 1

    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == expected_num_of_trophies

    # check that "made with love" text is not present
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.CLASS_NAME, "credits")


def test_dashboard__infinite_scroll(tracks_in_tmpdir, live_server, webdriver, insert_activity):
    rows_per_page = configuration.number_of_rows_per_page_in_table
    # insert more activities than the currently configured value of rows
    # per page in order to be  able to trigger the htmx ajax request
    nr_of_inserted_activities = rows_per_page + 5
    for i in range(nr_of_inserted_activities):
        insert_activity(name=f"Dummy Activity {i}")

    assert models.Activity.objects.count() == nr_of_inserted_activities
    webdriver.get(live_server.url + reverse("home"))

    # number of rows equals the number of rows per page, since only one page is loaded
    table_rows = [cell.text for cell in webdriver.find_elements_by_id("activities-table-row")]
    assert len(table_rows) + 1 == rows_per_page

    webdriver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
    # wait until loading image is present
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.ID, "loading-bar")))
    # wait until final row indicating that no more activities are available is present
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.ID, "end-of-activities")))

    # again check number of table rows
    table_rows = [cell.text for cell in webdriver.find_elements_by_id("activities-table-row")]
    assert len(table_rows) + 1 == nr_of_inserted_activities


def test_trophy_icon_for_awarded_activities_in_table_are_displayed_correctly(db, live_server, webdriver):
    assert models.Activity.objects.count() == 0

    # add a few activities
    sport = models.Sport.objects.create(name="Swimming", slug="swimming", icon="swimmer", color="#51CBCE")
    activity_1 = models.Activity.objects.create(
        name="Activity 1",
        sport=sport,
        date=datetime.datetime(2020, 1, 1, 15, 30, 0, tzinfo=pytz.UTC),
        duration=datetime.timedelta(0, 3600),
    )
    activity_2 = models.Activity.objects.create(
        name="Activity 2",
        sport=sport,
        date=datetime.datetime(2020, 1, 2, 15, 30, 0, tzinfo=pytz.UTC),
        duration=datetime.timedelta(0, 3600),
    )
    activity_3 = models.Activity.objects.create(
        name="Activity 3",
        sport=sport,
        date=datetime.datetime(2020, 1, 3, 15, 30, 0, tzinfo=pytz.UTC),
        duration=datetime.timedelta(0, 3600),
    )
    assert models.Sport.objects.count() == 1
    assert models.Activity.objects.count() == 3

    # verify no trophies are present by now
    webdriver.get(live_server.url + reverse("home"))
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 0

    # check that activities are already displayed in the table
    td = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    assert activity_1.name in td
    assert activity_2.name in td
    assert activity_3.name in td

    assert get_flat_list_of_pks_of_activities_in_top_awards() == []

    # now add best sections to activity 1 ...
    models.BestSection.objects.create(
        activity=activity_1,
        kind="fastest",
        distance=1000,
        start=42,
        end=128,
        max_value=123.45,
    )
    webdriver.get(live_server.url + reverse("home"))
    # ... and verify that 1 trophy is present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 1
    assert get_flat_list_of_pks_of_activities_in_top_awards() == [activity_1.pk]

    # add best section to activity 2 ...
    models.BestSection.objects.create(
        activity=activity_2,
        kind="fastest",
        distance=5000,
        start=43,
        end=129,
        max_value=99.99,
    )
    webdriver.get(live_server.url + reverse("home"))
    # ... and verify that 2 trophies are present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 2
    assert set(get_flat_list_of_pks_of_activities_in_top_awards()) == {activity_1.pk, activity_2.pk}

    # also add total ascent value to activity 3...
    trace = models.Traces.objects.create(
        path_to_file="dummy/path",
        file_name="foo.baa",
        md5sum="1a2b3c4d5e",
        total_ascent=500,
        total_descent=501,
    )
    activity_3.trace_file = trace
    activity_3.save()
    webdriver.get(live_server.url + reverse("home"))
    # ... and verify that 3 trophies are present
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) == 3
    assert get_flat_list_of_pks_of_activities_in_top_awards() == [activity_1.pk, activity_2.pk, activity_3.pk]



import os

import setuptools

from workoutizer import __version__

with open("Readme.md", "r") as fh:
    long_description = fh.read()


def requirements_from_txt(path_to_txt):
    path_to_reqs = os.path.join(os.path.dirname(__file__), "setup", "requirements")
    with open(os.path.join(path_to_reqs, path_to_txt), "r") as f:
        reqs = f.readlines()
    return [req for req in reqs if not req.startswith("#")]


setuptools.setup(
    name="workoutizer",
    author="Fabian Gebhart",
    version=__version__,
    setup_requires=["setuptools_scm"],
    install_requires=requirements_from_txt("requirements.txt"),
    include_package_data=True,
    description="🏋️ Browser based Sport and Workout Organizer 🏃‍♀️",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fgebhart/workoutizer",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
    ],
    python_requires=">=3.7",
    extras_require={
        "testing": requirements_from_txt("dev-requirements.txt"),
    },
    packages=setuptools.find_packages(exclude=["tests"]) + ["setup", "tracks", "media"],
    entry_points={"console_scripts": ["wkz = workoutizer.cli:wkz"]},
)
